<?php
/*
 * GREYHOUND Addon Framework
 *
 * Copyright (c) 2011-2019 GREYHOUND Software GmbH & Co. KG
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of the GREYHOUND Addon Framework and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including without
 * limitation the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to permit persons to
 * whom the Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/**
 * Addon-Klasse für GREYHOUND Addons.
 *
 * @package GhAddonFramework
 * @subpackage core
 * @category PHP Addon Framework
 *
 * @author digital guru GmbH &amp; Co. KG <develop@greyhound-software.com>
 * @copyright 2011-2019 digital guru GmbH &amp; Co. KG
 * @link greyhound-software.com
 */
abstract class GhAddon
{
  /**
   * URL des Addon Frameworks.
   * @var string
   */
  protected static $_frameworkUrl = 'lib/ghAddonFramework/';

  /**
   * Name des Addons bzw. des Drittanbieters.
   * @var string
   */
  protected $_addonName;

  /**
   * Name des Systems des Drittanbieters.
   * @var string
   */
  protected $_remoteServiceName;

  /**
   * Prefix für den Identifier von Extensions.
   * @var string
   */
  protected $_extensionIdentifierPrefix;

  /**
   * Versionsnummer des Addons.
   * @see $_addonBuild
   * @var string
   */
  protected $_addonVersion = '0.0.1';

  /**
   * Build-Nummer des Addons.
   * Die Build-Nummer wird unter anderem an die URL für eingebundene Javascript
   * Dateien als GET-Parameter angehängt, um sicherzustellen, dass bei neuen
   * Versionen die Scripte neu abgerufen werden (und nicht aus dem Cache
   * geladen werden).
   * @see $_addonVersion
   * @var string
   */
  protected $_addonBuild = 0;

  /**
   * Icon-URL/Pfad des Addons.
   * @var string
   */
  protected $_addonIcon = null;

  /**
   * Tooltip-Text für das Suchfeld.
   * @var string
   */
  protected $_searchTooltip = '';

  /**
   * Name/Identifier des benutzerdefinierten Felds für die Kunden-ID des Drittanbieters.
   * @var string
   */
  protected $_customerIdUserFieldName;

  /**
   * Beschreibung des benutzerdefinierten Felds für die Kunden-ID des Drittanbieters.
   * @var string
   */
  protected $_customerIdUserFieldDescription;

  /**
   * Typ des benutzerdefinierten Felds für die Kunden-ID des Drittanbieters.
   * @var int
   */
  protected $_customerIdUserFieldType = rpc_userfieldslib::ufdtString;

  /**
   * E-Mail Adresse des Maintainers des Addons.
   * @var string
   */
  protected $_maintainerEmail = 'develop@greyhound-software.com';

  /**
   * Name des Maintainers des Addons.
   * @var string
   */
  protected $_maintainerName = 'GREYHOUND Entwicklung';

  /**
   * Produkt-ID des Addons.
   * Diese ID wird verwendet um die Einstellungen, die auf dem GREYHOUND Server
   * gespeichert werden, zu identifizieren.
   * @var string
   */
  protected $_addonProductId;

  /**
   * Name der Session-ID für dieses Addon.
   * @var string
   */
  protected $_sessionName;

  /**
   * Accounts.
   * Die Accounts werden zur Laufzeit in dieser Variablen gecached.
   * @var array
   */
  protected $_accounts = false;

  /**
   * GREYHOUND Client-Objekt (für die Verbindung zum GREYHOUND Server).
   * @var GhRpcClient
   */
  protected $_ghClient = false;

  /**
   * GREYHOUND UserInfo für den RPC Benutzer.
   * Die UserInfo wird zur Laufzeit in dieser Variablen gecached.
   * @var GhRpcUserslibRpcUserInfo
   */
  protected $_ghUserInfo = false;

  /**
   * Cache-Timeout (in Sekunden) für Addon-Einstellungen.
   * @var int
   */
  protected $_ghUserInfoCacheTimeout = 300; // 5 Minuten

  /**
   * Einstellungen des Addons.
   * Die Einstellungen werden zur Laufzeit in dieser Variablen gecached.
   * @var array
   */
  protected $_addonSettings = false;

  /**
   * Cache-Timeout (in Sekunden) für Addon-Einstellungen.
   * @var int
   */
  protected $_addonSettingsCacheTimeout = 300; // 5 Minuten

  /**
   * Filter für Empfänger.
   * Die Empfängerfilter werden zur Laufzeit in dieser Variablen gecached.
   * @var array
   */
  protected $_recipientFilters = false;

  /**
   * Cache-Timeout (in Sekunden) für Empfängerfilter.
   * @var int
   */
  protected $_recipientFiltersCacheTimeout = 300; // 5 Minuten

  /**
   * True falls sich die Kunden-ID und Kundennummer unterscheiden können, false, falls sie immer identisch sind (es
   * also keine getrennten Felder für Kunden-ID und Kundennummer gibt).
   * @var boolean
   */
  protected $_customerIdAndNumberDiffer = false;


  /**
   * Setzt die URL des Addon Frameworks.
   * @param string $url Addon Framework URL
   */
  public static function setAddonFrameworkUrl($url)
  {
    GhAddon::$_frameworkUrl = $url;
  }

  /**
   * Liefert die URL des Addon Frameworks.
   *
   * @return string
   */
  public static function getAddonFrameworkUrl()
  {
    return GhAddon::$_frameworkUrl;
  }

  /**
   * Konstruktor.
   */
  public function __construct()
  {
    // Session weiterführen (oder anlegen):
    GhCoreGlobals::continueSession($this->getSessionName());
  }

  /**
   * Liefert den Namen der Session-ID für dieses Addon.
   *
   * @return string
   */
  public function getSessionName()
  {
    if(!$this->_sessionName)
      $this->_sessionName = get_class($this);

    return $this->_sessionName;
  }

  /**
   * Diese Methode liefert den Namen einer Addon-Einstellung, anhand derer sich
   * abzählen lässt, wieviele Accounts existieren.
   * Dazu wird geprüft, wieviele Einstellungen existieren, die mit diesem Namen
   * beginnen und mit einer Ziffer enden.
   *
   * @return string
   */
  public function getAccountMainSettingName()
  {
    return 'Caption';
  }

  /**
   * Diese Methode muss überladen werden und einen Account für das Addon
   * liefern.
   *
   * @param int   $id       Account-ID
   * @param array $settings Account-Einstellungen
   * @return GhAddonAccount
   */
  abstract protected function _createAccount($id, $settings);

  /**
   * Prüft Verbindungsdaten und versucht, eine Verbindung zum Drittsystem herzustellen.
   *
   * @throws GhAddonConnectionException falls keine Verbindung hergestellt werden konnte
   *
   * @param array $data Account-Einstellungen für die Verbindung
   * @return boolean
   */
  abstract public function checkConnection($data);

  /**
   * Liefert den Titel des Addons bzw. des Drittanbieters.
   *
   * @return string
   */
  public function getName()
  {
    return $this->_addonName;
  }

  /**
   * Liefert den Namen des Drittsystems.
   *
   * @return string
   */
  public function getRemoteServiceName()
  {
    return $this->_remoteServiceName;
  }

  /**
   * Liefert den Tooltip-Text für das Suchfeld.
   *
   * @return string
   */
  public function getSearchTooltip()
  {
    return $this->_searchTooltip;
  }

  /**
   * Liefert den Anfang des Bezeichners der Erweiterungen dieses Addons.
   * @return string
   */
  protected function _getExtensionIdentifierPrefix()
  {
    return $this->_extensionIdentifierPrefix ? $this->_extensionIdentifierPrefix : $this->_addonName;
  }

  /**
   * Liefert den Bezeichner der Seitenleisten-Erweiterung.
   *
   * @return string
   */
  public function getSidebarIdentifier()
  {
    return $this->_getExtensionIdentifierPrefix() . 'Sidebar';
  }

  /**
   * Liefert den Bezeichner der Einstellungsdialog-Erweiterung.
   * @return string
   */
  public function getSettingsIdentifier()
  {
    return $this->_getExtensionIdentifierPrefix() . 'Settings';
  }

  /**
   * Liefert den Bezeichner der Webbrowser-Erweiterung.
   * @return string
   */
  public function getWebIdentifier()
  {
    return $this->_getExtensionIdentifierPrefix() . 'WebLink';
  }

  /**
   * Liefert den Bezeichner der Hilfe-Erweiterung.
   * @return string
   */
  public function getHelpIdentifier()
  {
    return $this->_getExtensionIdentifierPrefix() . 'Help';
  }

  /**
   * Liefert die E-Mail Adresse des Maintainers des Addons.
   * Diese wird von der Debug-Funktion genutzt um z.B. Daten eines Requests
   * zu Debugging-Zwecken an den Maintainer senden zu können.
   *
   * @return string
   */
  public function getMaintainerEmail()
  {
    if(!$this->_maintainerEmail)
      return null;

    if($this->_maintainerName)
      return $this->_maintainerName . ' <' . $this->_maintainerEmail . '>';
    else
      return '<' . $this->_maintainerEmail . '>';
  }

  /**
   * Liefert die Version des Addons.
   *
   * @return string
   */
  public function getVersion()
  {
    return $this->_addonVersion;
  }

  /**
   * Liefert die Build-Nummer des Addons.
   *
   * @return int
   */
  public function getBuild()
  {
    return $this->_addonBuild;
  }

  /**
   * Liefert die URL bzw. den relativen Pfad zum Icon des Addons.
   *
   * @return string
   */
  public function getIcon()
  {
    return $this->_addonIcon;
  }

  /**
   * Liefert alle Accounts des Addons.
   *
   * @return GhAddonAccount[]
   */
  public function getAccounts()
  {
    if(!is_array($this->_accounts))
    {
      $this->_accounts = array();

      foreach($this->getAddonAccountSettings(true) as $id => $settings)
        $this->_accounts[$id] = $this->_createAccount($id, $settings);
    }

    return $this->_accounts;
  }

  /**
   * Liefert einen Account anhand seiner ID.
   *
   * @param int $id Account-ID
   * @return GhAddonAccount
   */
  public function getAccount($id)
  {
    $accounts = $this->getAccounts();

    if(isset($accounts[$id]))
      return $accounts[$id];
    else
      return null;
  }

  /**
   * Liefert das GREYHOUND Client Objekt (für die Verbindung zum GREYHOUND Server).
   *
   * @return GhRpcClient
   */
  public function getGreyhoundClient()
  {
    if($this->_ghClient === false)
    {
      $ghHost = GhCoreGlobals::getValue('ghhost', null, 'GPS');
      $ghPort = GhCoreGlobals::getValue('ghport', null, 'GPS');
      $ghProtocol = GhCoreGlobals::getValue('ghprotocol', null, 'GPS');
      $ghToken = GhCoreGlobals::getValue('ghtoken', null, 'GPS');
      $ghUniqueClientId = GhCoreGlobals::getValue('ghclientid', null, 'GPS');

      GhCoreGlobals::session('ghhost', $ghHost, true);
      GhCoreGlobals::session('ghport', $ghPort, true);
      GhCoreGlobals::session('ghprotocol', $ghProtocol, true);
      GhCoreGlobals::session('ghtoken', $ghToken, true);
      GhCoreGlobals::session('ghclientid', $ghUniqueClientId, true);

      $this->_ghClient = new GhRpcClient($ghHost, $ghPort, $ghProtocol == 'https', 'token', $ghToken);
      $this->_ghClient->setUniqueClientIdCookieName($this->_getGreyhoundClientUniqueIdCookieName());

      if ($ghUniqueClientId) {
        $this->_ghClient->setUniqueClientId($ghUniqueClientId);
      }
    }

    return $this->_ghClient;
  }

  /**
   * Liefert den Namen des Cookies für die eindeutige Client-ID des
   * Client-Objekts.
   * Da alle Addons in der Regel im selben Client laufen, verwenden sie
   * denselben Cookie.
   *
   * @return string
   */
  protected function _getGreyhoundClientUniqueIdCookieName()
  {
    return 'ghAddon';
  }

  /**
   * Liefert die UserInfo vom GREYHOUND Server für den angemeldeten Benutzer.
   *
   * @throws GhRpcException falls ein Fehler beim Abrufen der UserInfo vom GREYHOUND Server auftritt (und $throwException true ist)
   *
   * @param boolean $refreshCache   true: Daten nicht aus dem Cache verwenden, Cache ggf. erneuern, false: Daten aus dem Cache verwenden, falls vorhanden
   * @param boolean $throwException true: RPC Exceptions werden durchgelassen, false: RPC Exceptions werden ignoriert
   * @return GhRpcUserslibRpcUserInfo
   */
  public function getGreyhoundUserInfo($refreshCache = false, $throwException = false)
  {
    // Laufzeit-Cache prüfen:
    if(!$refreshCache && $this->_ghUserInfo !== false)
      return $this->_ghUserInfo;

    // Session-Cache prüfen (und ggf. in Laufzeit-Cache übernehmen):
    if(!$refreshCache)
    {
      $sessionTime = (int)GhCoreGlobals::session('userInfoTime');

      if($sessionTime < time() - $this->_ghUserInfoCacheTimeout)
        $this->_clearGreyhoundUserInfoCache();
      else
      {
        $this->_ghUserInfo = $this->_readGreyhoundUserInfoFromSession();

        if(!$this->_ghUserInfo)
          $this->_ghUserInfo = null;

        return $this->_ghUserInfo;
      }
    }

    // UserInfo vom Server abrufen und in Laufzeit- und Session-Cache übernehmen:
    try
    {
      $this->_ghUserInfo = $this->_readGreyhoundUserInfoFromGreyhound();
    }
    catch(GhRpcException $exception)
    {
      if($throwException)
        throw $exception;
    }

    if(!$this->_ghUserInfo)
      $this->_ghUserInfo = null;

    // Einstellungen im Session-Cache ablegen:
    $this->_writeGreyhoundUserInfoToSession($this->_ghUserInfo);

    return $this->_ghUserInfo;
  }

  /**
   * Liefert die GREYHOUND UserInfo aus dem Session-Cache.
   *
   * @return array
   */
  protected function _readGreyhoundUserInfoFromSession()
  {
    $sessionData = GhCoreGlobals::session('userInfo');

    if($sessionData)
    {
      rpc_userslib::newRpcUserInfo();// Klasse vom Autoloader laden lassen
      return unserialize($sessionData);
    }
    else
      return null;
  }

  /**
   * Speichert die GREYHOUND UserInfo im Session-Cache.
   *
   * @param GhRpcUserslibRpcUserInfo $userInfo UserInfo
   */
  protected function _writeGreyhoundUserInfoToSession($userInfo)
  {
    GhCoreGlobals::session('userInfo', serialize($userInfo), true);
    GhCoreGlobals::session('userInfoTime', time(), true);
  }

  /**
   * Leert den Laufzeit- und Session-Cache für die GREYHOUND UserInfo.
   */
  protected function _clearGreyhoundUserInfoCache()
  {
    $this->_ghUserInfo = false;
    GhCoreGlobals::session('userInfo', null, true);
    GhCoreGlobals::session('userInfoTime', null, true);
  }

  /**
   * Liefert die UserInfo vom GREYHOUND Server.
   *
   * @throws GhRpcException falls beim Abrufen der UserInfo vom GREYHOUND Server ein Fehler aufgetreten ist
   *
   * @return array
   */
  protected function _readGreyhoundUserInfoFromGreyhound()
  {
    return rpc_userslib::newRpcUsers($this->getGreyhoundClient())->GetUserInfo();
  }

  /**
   * Prüft, ob eine Produkt-ID für dieses Addon festgelegt wurde.
   *
   * @throws GhAddonException falls keine Produkt-ID für dieses Addon festgelegt wurde (in der Variablen $_addonProductId)
   */
  protected function _assertProductId()
  {
    if(!$this->_addonProductId)
      throw new GhAddonException('No product id has been specified in the GhAddon class of this addon.');
  }

  /**
   * Liefert die Einstellungen die für das Addon auf dem GREYHOUND Server gespeichert wurden.
   *
   * @throws GhAddonException falls keine Produkt-ID für dieses Addon festgelegt wurde (in der Variablen $_addonProductId)
   * @throws GhRpcException falls beim Abrufen der Einstellungen vom GREYHOUND Server ein Fehler aufgetreten ist
   *
   * @param boolean $refreshCache true: Daten nicht aus dem Cache verwenden, Cache ggf. erneuern, false: Daten aus dem Cache verwenden, falls vorhanden
   * @return array
   */
  public function getAddonSettings($refreshCache = false)
  {
    $this->_assertProductId();

    // Laufzeit-Cache prüfen:
    if(!$refreshCache && is_array($this->_addonSettings))
      return $this->_addonSettings;

    // Session-Cache prüfen (und ggf. in Laufzeit-Cache übernehmen):
    if(!$refreshCache)
    {
      $sessionTime = (int)GhCoreGlobals::session('addonSettingsTime');

      if($sessionTime < time() - $this->_addonSettingsCacheTimeout)
        $this->_clearAddonSettingsCache();
      else
      {
        $this->_addonSettings = $this->_readAddonSettingsFromSession();

        if(!is_array($this->_addonSettings))
          $this->_addonSettings = array();

        $this->_addonSettings = $this->_processAddonSettingsAfterLoad($this->_addonSettings);

        return $this->_addonSettings;
      }
    }

    // Einstellungen vom Server abrufen und in Laufzeit- und Session-Cache übernehmen:
    $this->_addonSettings = $this->_readAddonSettingsFromGreyhound();

    if(!$this->_addonSettings)
      $this->_addonSettings = array();

    // Standard-Setup für die Vorschauzeile:
    if(!isset($this->_addonSettings['PreviewLineField.0.Var']))
    {
      foreach($this->_getPreviewLineDefaultSettings() as $key => $value)
        $this->_addonSettings[$key] = $value;
    }

    // Felder aus der Vorschauzeile entfernen, die das Addon nicht anbietet:
    $addonVars = array_keys($this->getTemplateVariableDescriptions());
    $previewLineFields = array();
    $i = 0;

    while(isset($this->_addonSettings['PreviewLineField.' . $i . '.Var']))
    {
      if(in_array($this->_addonSettings['PreviewLineField.' . $i . '.Var'], $addonVars))
        $previewLineFields[] = array('Var' => $this->_addonSettings['PreviewLineField.' . $i . '.Var'], 'Caption' => $this->_addonSettings['PreviewLineField.' . $i . '.Caption']);

      unset($this->_addonSettings['PreviewLineField.' . $i . '.Var']);
      unset($this->_addonSettings['PreviewLineField.' . $i . '.Caption']);
      $i++;
    }

    for($i = 0; $i < count($previewLineFields); $i++)
    {
      $this->_addonSettings['PreviewLineField.' . $i . '.Var'] = $previewLineFields[$i]['Var'];
      $this->_addonSettings['PreviewLineField.' . $i . '.Caption'] = $previewLineFields[$i]['Caption'];
    }

    // Ggf. überladene Funktion für die Nachbehandlung der Settings aufrufen:
    $this->_addonSettings = $this->_processAddonSettingsAfterLoad($this->_addonSettings);

    // Einstellungen im Session-Cache ablegen:
    $this->_writeAddonSettingsToSession($this->_addonSettings);

    return $this->_addonSettings;
  }

  /**
   * Liefert die Benutzer-Einstellungen des aktuellen Benutzers für das Addon.
   * 
   * @return array
   */
  public function getAddonUserSettings()
  {
    $this->_assertProductId();

    $settings = rpc_userslib::newRpcUsers($this->getGreyhoundClient())->GetSettings($this->_addonProductId);

    if ($settings) {
      try {
        return json_decode($settings);
      } catch (Exception $e) {
        return (object)array();
      }
    } else {
      return (object)array();
    }
  }

  /**
   * Setzt die Benutzer-Einstellungen des aktuellen Benutzers für das Addon.
   * 
   * @param array $settings Benutzer-Einstellungen
   */
  public function setAddonUserSettings($settings)
  {
    $this->_assertProductId();

    rpc_userslib::newRpcUsers($this->getGreyhoundClient())->PutSettings($this->_addonProductId, json_encode($settings));
  }

  /**
   * Setzt eine Benutzer-Einstellung des aktuellen Benutzers für das Addon.
   * 
   * @param string $key Name der Einstellung
   * @param mixed $value Wert der Einstellung
   */
  public function setAddonUserSetting($key, $value)
  {
    $settings = $this->getAddonUserSettings();

    if (strlen($key)) {
      $path = explode('/', $key);
      $last = array_pop($path);
      $obj = &$settings;

      foreach ($path as $part) {
        if (!isset($obj->$part)) {
          $obj->$part = (object)array();
        }

        $prop = $obj->part;
        $obj = &$prop;
      }

      $obj->$last = $value;
    }

    $this->setAddonUserSettings($settings);
  }

  /**
   * Liefert die Voreinstellungen für die Vorschauzeile.
   * 
   * @return array
   */
  protected function _getPreviewLineDefaultSettings()
  {
    return array(
      'PreviewLineField.0.Var' => 'Order.OrderNumber',
      'PreviewLineField.0.Caption' => 'Auftrag:',
      'PreviewLineField.1.Var' => 'Order.Status',
      'PreviewLineField.1.Caption' => 'Status:',
      'PreviewLineField.2.Var' => 'Order.CustomerNumber',
      'PreviewLineField.2.Caption' => 'Kunde:',
      'PreviewLineField.3.Var' => 'Order.TotalAmount',
      'PreviewLineField.3.Caption' => 'Summe:',
      'PreviewLineField.4.Var' => 'Payment.Method',
      'PreviewLineField.4.Caption' => 'Zahlung:',
      'PreviewLineField.5.Var' => 'Order.OpenAmount',
      'PreviewLineField.5.Caption' => 'Offen:',
      'PreviewLineField.6.Var' => 'Payment.Date',
      'PreviewLineField.6.Caption' => '',
      'PreviewLineField.7.Var' => 'Shipping.Provider',
      'PreviewLineField.7.Caption' => 'Versand:',
      'PreviewLineField.8.Var' => 'Shipping.Costs',
      'PreviewLineField.8.Caption' => '',
      'PreviewLineField.9.Var' => 'Shipping.Tracking.URL',
      'PreviewLineField.9.Caption' => '',
      'PreviewLineField.10.Var' => 'Order.InvoiceNumber',
      'PreviewLineField.10.Caption' => 'Letzte Rechnung:'
    );
  }

  /**
   * Speichert die Einstellungen für das Addon.
   *
   * @throws GhAddonException falls keine Produkt-ID für dieses Addon festgelegt wurde (in der Variablen $_addonProductId)
   * @throws GhRpcException falls beim Speichern der Einstellungen im GREYHOUND Server ein Fehler aufgetreten ist
   *
   * @param array $settings Einstellungen
   */
  public function setAddonSettings($settings)
  {
    $this->_assertProductId();

    $settings = $this->_processAddonSettingsBeforeSave($settings);

    $this->_writeAddonSettingsToGreyhound($settings);
    $this->_clearAddonSettingsCache();

    // Laufzeit-Cache zurücksetzen:
    $this->_addonSettings = false;
    $this->_accounts = false;

    // Cache der Accounts leeren:
    foreach ($this->getAccounts() as $account) {
      $account->clearCache();
    }
  }

  /**
   * Liefert die Einstellungen des Addons aus dem Session-Cache.
   *
   * @return array
   */
  protected function _readAddonSettingsFromSession()
  {
    $this->_assertProductId();

    $sessionSettings = GhCoreGlobals::session('addonSettings');

    if($sessionSettings)
      return unserialize($sessionSettings);
    else
      return null;
  }

  /**
   * Speichert die Einstellungen des Addons im Session-Cache.
   *
   * @param array $settings Addon-Einstellungen
   */
  protected function _writeAddonSettingsToSession($settings)
  {
    $this->_assertProductId();

    GhCoreGlobals::session('addonSettings', serialize($settings), true);
    GhCoreGlobals::session('addonSettingsTime', time(), true);
  }

  /**
   * Leert den Laufzeit- und Session-Cache für die Addon-Einstellungen.
   */
  protected function _clearAddonSettingsCache()
  {
    $this->_assertProductId();

    $this->_addonSettings = false;
    GhCoreGlobals::session('addonSettings', null, true);
    GhCoreGlobals::session('addonSettingsTime', null, true);
  }

  /**
   * Liefert die Einstellungen des Addons vom GREYHOUND Server.
   *
   * @throws GhAddonException falls keine Produkt-ID für dieses Addon festgelegt wurde (in der Variablen $_addonProductId)
   * @throws GhRpcException falls beim Abrufen der Einstellungen vom GREYHOUND Server ein Fehler aufgetreten ist
   *
   * @return array
   */
  protected function _readAddonSettingsFromGreyhound()
  {
    $this->_assertProductId();

    $settings = array();

    $settingsText = rpc_extensionslib::newRpcExtensions($this->getGreyhoundClient())->GetSettings($this->_addonProductId);

    foreach(explode("\n", $settingsText) as $line)
    {
      $split = strpos($line, '=');

      if($split !== false)
        $settings[substr($line, 0, $split)] = substr($line, $split + 1);
    }

    return $settings;
  }

  /**
   * Speichert die Einstellungen des Addons auf dem GREYHOUND Server.
   *
   * @throws GhAddonException falls keine Produkt-ID für dieses Addon festgelegt wurde (in der Variablen $_addonProductId)
   * @throws GhRpcException falls beim Speichern der Einstellungen im GREYHOUND Server ein Fehler aufgetreten ist
   */
  protected function _writeAddonSettingsToGreyhound($settings)
  {
    $this->_assertProductId();

    $settingsText = '';

    foreach($settings as $key => $value)
    {
      $settingsText .= $key . '=' . $value . "\n";
    }

    rpc_extensionslib::newRpcExtensions($this->getGreyhoundClient())->PutSettings($this->_addonProductId, $settingsText);
  }

  /**
   * Liefert die Account-Einstellungen des Addons zurück.
   *
   * Das Ergebnis ist ein Array, dessen Schlüssel die IDs der Accounts sind
   * (durchnummeriert, beginnend mit 0) und dessen Werte Arrays mit
   * Schlüssel-Wert Paaren für die Einstellungen des jeweiligen Accounts sind.
   *
   * @param boolean $stripAccountId true: Präfix und Account-ID aus den Einstellungs-Namen entfernen, false: Namen der Einstellungen unverändert lassen
   * @param array   $settings       Array mit Einstellungen, der in Account-Einstellungen sortiert werden soll. Wird kein Array angegeben,
   *   so werden die Einstellungen vom GREYHOUND Server abgerufen.
   * @return array
   */
  public function getAddonAccountSettings($stripAccountId = true, $settings = null)
  {
    $accounts = array();

    if(!is_array($settings))
      $settings = $this->getAddonSettings();

    foreach($settings as $setting => $value)
    {
      $matches = array();

      if(preg_match_all('/^Account\.([0-9]+)\.(.*)$/', $setting, $matches, PREG_SET_ORDER) && count($matches) > 0)
      {
        $accountId = $matches[0][1];

        if($stripAccountId)
          $setting = $matches[0][2];

        if(!array_key_exists($accountId, $accounts))
          $accounts[$accountId] = array($setting => $value);
        else
          $accounts[$accountId][$setting] = $value;
      }
    }

    // Accounts entfernen, die nicht über die Haupt-Einstellung verfügen:
    $mainSetting = $this->getAccountMainSettingName();

    foreach($accounts as $accountId => $accountSettings)
      if(($stripAccountId && !isset($accountSettings[$mainSetting])) || (!$stripAccountId && !isset($accountSettings['Account.' . $accountId . '.' . $mainSetting])))
        unset($accounts[$accountId]);

    return $accounts;
  }

  /**
   * Liefert die globalen Einstellungen des Addons zurück.
   *
   * Das Ergebnis ist ein Array mit Schlüssel-Wert Paaren.
   *
   * @param array $settings Array mit Einstellungen, aus denen die globalen
   *   Einstellungen extrahiert werden sollen. Wird kein Array angegeben, so
   *   werden die Einstellungen vom GREYHOUND Server (oder aus dem Cache)
   *   abgerufen.
   * @return array
   */
  public function getAddonGlobalSettings($settings = null)
  {
    $globalSettings = array();

    if(!is_array($settings))
      $settings = $this->getAddonSettings();

    foreach($settings as $name => $value)
    {
      if(!$this->isAddonSettingGlobal($name))
        continue;

      $globalSettings[$name] = $value;
    }

    return $globalSettings;
  }

  /**
   * Prüft, ob eine Einstellung global oder Account-spezifisch ist.
   *
   * Ist die Einstellung global, so wird true zurück geliefert, ist sie
   * Account-spezifisch, so wird false zurück geliefert.
   *
   * @param string $name Name der Einstellung
   * @return boolean
   */
  public function isAddonSettingGlobal($name)
  {
    if(preg_match_all('/^Account\.[0-9]+\..*$/', $name))
      return false;
    else
      return true;
  }

  /**
   * Liefert den Wert einer Einstellung, die für dieses Addon auf dem GREYHOUND Server gespeichert wurde.
   *
   * @throws GhAddonException falls keine Produkt-ID für dieses Addon festgelegt wurde (in der Variablen $_addonProductId)
   *
   * @param string $name Name der Einstellung
   * @param string $defaultValue Vorgabewert, falls die Einstellung nicht existiert
   * @return string
   */
  public function getAddonSetting($name, $defaultValue = null)
  {
    $settings = $this->getAddonSettings();

    if(is_array($settings) && isset($settings[$name]))
      return $settings[$name];
    else
      return $defaultValue;
  }

  /**
   * Liefert einen Account, der zu den übergebenen Empfängern passt.
   *
   * Passen die Empfänger mehrerer Accounts, so wird der erste zurück geliefert.
   * Gibt es keinen passenden Account, so wird null zurück geliefert.
   *
   * @param array|string $recipients
   * @return int
   */
  public function getAccountByRecipient($recipients = null)
  {
    foreach($this->getAccounts() as $account)
    {
      /* @var GhAddonAccount $account */
      if($account->isValid() && $account->isActive() && $account->isValidRecipient($recipients))
        return $account;
    }

    return null;
  }

  /**
   * Liefert den Account für die Verarbeitung einer Anfrage oder Suche.
   *
   * @return GhAddonAccount
   */
  public function getAccountByRequest()
  {
    $itemState = GhCoreGlobals::getValue('itemState');
    $itemCc = GhCoreGlobals::getValue('itemCC');
    $recipient = trim(GhCoreGlobals::getValue('itemRecipient'));
    $searchTerm = trim(GhCoreGlobals::getValue('search'));

    $recipients = array();

    if($itemState == rpc_itemslib::isNew || $itemState == rpc_itemslib::isAnswer || $itemState == rpc_itemslib::isForward || $itemState == rpc_itemslib::isDraft)
      $recipients[] = $this->getEmailAddressByRequest();
    else
    {
      if($recipient)
        $recipients[] = $recipient;

      if(is_array($itemCc))
        foreach($itemCc as $tmpCc)
          $recipients[] = trim($tmpCc);
    }

    $account = $this->getAccountByRecipient($recipients);

    // Bei einer Suche wird der erste aktive Account verwendet, wenn sonst kein Account zum Empfänger passt:
    if(!$account && $searchTerm)
    {
      foreach($this->getAccounts() as $tmpAccount)
      {
        /* @var GhAddonAccount $tmpAccount */
        if($tmpAccount->isValid() && $tmpAccount->isActive())
        {
          $account = $tmpAccount;
          break;
        }
      }
    }

    return $account;
  }

  /**
   * Liefert die E-Mail Adresse für die Verarbeitung einer Anfrage oder Suche.
   *
   * @return string
   */
  public function getEmailAddressByRequest()
  {
    $email = GhCoreGlobals::getValue('itemReplyToEmail') ? GhCoreGlobals::getValue('itemReplyToEmail') : GhCoreGlobals::getValue('itemEmail');
    $emailFallback = GhCoreGlobals::getValue('itemEmailFallback');

    // Service-E-Mail Adressen anullieren (interessant ist nur die Kunden-Adresse):
    if($email == 'service@paypal.de' || strpos($email, '@members.ebay.') > 0) {
      $email = '';
    }

    // Amazon-Marketplace E-Mail Adressen bei Bedarf bereinigen:
    $index = strpos($email, '@marketplace.amazon.');

    if($index > 0) {
      $name = substr($email, 0, $index);
      $domain = substr($email, $index);
      $index2 = strrpos($name, '+');

      if($index2 > 0) {
        $email = substr($name, 0, $index2) . $domain;
      }
    }

    // Falls keine E-Mail Adresse bzw. kein eBay-Name vorliegt wird als Fallback auf einen regulären Ausdruck über den Inhalt zurück gegriffen:
    if(!$email && $emailFallback && strpos($emailFallback, '@') !== false) {
      $email = $emailFallback;
    }

    return $email;
  }

  /**
   * Ändert die Vorgangsnummer eines GREYHOUND Elements.
   *
   * @throws Exception falls beim Setzen der Vorgangsnummer ein Fehler auftritt
   *
   * @param int    $itemId     Element-ID
   * @param string $taskNumber neue Vorgangsnummer
   * @param string $remark     Anmerkung
   */
  public function setItemTaskNumber($itemId, $taskNumber, $remark = '')
  {
    $lockData = null;
    $itemService = rpc_itemslib::newRpcItems($this->getGreyhoundClient());

    try {
      $lockData = $itemService->GetLockData($itemId);

      if (!$lockData)
        $itemService->Lock($itemId);

      $itemService->SetNumber($itemId, $taskNumber, $remark);

      if (!$lockData)
        $itemService->Unlock($itemId);
    } catch (Exception $exception) {
      // Sicherstellen, dass das Element bei einem Fehler nicht gesperrt bleibt:
      try {
        if (!$lockData)
          $itemService->Unlock($itemId);
      } catch (Exception $exception2) {
      }

      throw $exception;
    }
  }

  /**
   * Liefert die Signatur einer Gruppe.
   *
   * @param int    $groupId Gruppen-ID
   * @param string $format  Signatur-Format ('html' oder 'text')
   * @return string
   */
  public function getSignature($groupId, $format = 'html')
  {
    try
    {
      $group = rpc_groupslib::newRpcGroups($this->getGreyhoundClient())->Get($groupId);

      if(!$group || !$group->SignatureRef)
        return '';

      $signature = rpc_signatureslib::newRpcSignatures($this->getGreyhoundClient())->Get($group->SignatureRef);

      if(!$signature)
        return '';

      if($format == 'html' && $signature->HtmlData)
        return $signature->HtmlData;
      elseif($format == 'text' && $signature->TextData)
        return $signature->TextData;
      else
        return '';
    }
    catch(Exception $exception)
    {
      return '';
    }
  }

  /**
   * Liefert das Format für neue/beantwortete/weitergeleitete Elemente (E-Mails, Faxe, Briefe).
   *
   * @param int     $itemKind       Element-Typ
   * @param int     $itemState      Element-Status
   * @param boolean $isTemplateHtml true falls das Element mit einer
   *   HTML-Vorlage erzeugt werden soll, false falls das Element mit einer
   *   Text-Vorlage erzeugt werden soll, null andernfalls (falls das Element
   *   ohne Vorlage erzeugt werden soll)
   * @return string
   */
  public function getItemFormat($itemKind = rpc_itemslib::ikEmail, $itemState = rpc_itemslib::insNew, $isTemplateHtml = null)
  {
    $userInfo = $this->getGreyhoundUserInfo();

    if(isset($userInfo->ItemFormats[$itemKind]))
    {
      $formats = $userInfo->ItemFormats[$itemKind];

      if(is_object($formats))
        $formats = get_object_vars($formats);

      switch($itemState)
      {
        case rpc_itemslib::insNew:
          $formatKey = 'New';
          break;

        case rpc_itemslib::insAnswer:
          $formatKey = 'Answer';
          break;

        case rpc_itemslib::insForward:
          $formatKey = 'Forward';
          break;

        default:
          $formatKey = '';
          break;
      }

      if($formatKey)
      {
        if(!$formats || !$formats[$formatKey] || $formats[$formatKey] == rpc_profileslib::fnText)
          return 'text';
        elseif($formats[$formatKey] == rpc_profileslib::fnHtml)
          return 'html';
      }
    }

    if($isTemplateHtml !== null)
      return $isTemplateHtml ? 'html' : 'text';

    return 'html';
  }

  /**
   * Liefert die anzuzeigenden Einstellungs-Seiten.
   * 
   * @return array
   */
  public function getOptionPages()
  {
    return array('connect', 'global', 'migration', 'preview', 'tracking', 'extended');
  }

  /**
   * Liefert die Account-Einstellungs-Optionen des Addons.
   *
   * @param array $settings  aktuelle Einstellungen des Addons
   * @param int   $accountId aktuelle Account-ID
   * @return array
   */
  public function getAccountOptions($settings = null, $accountId = null)
  {
    return array();
  }

  /**
   * Liefert die globalen Einstellungs-Optionen des Addons.
   *
   * @param array $settings aktuelle Einstellungen des Addons
   * @return array
   */
  public function getGlobalOptions($settings = null)
  {
    return array();
  }

  /**
   * Liefert die erweiterten Einstellungs-Optionen des Addons.
   *
   * @param array $settings aktuelle Einstellungen des Addons
   * @return array
   */
  public function getExtendedOptions($settings = null)
  {
    return array();
  }

  /**
   * Liefert den Hilfetext des Addons.
   * 
   * @return string
   */
  public function getHelp()
  {
    return '';
  }

  /**
   * Modifiziert die Addon-Einstellungen nachdem sie geladen wurden.
   *
   * Diese Methode kann überladen werden, um nach dem Laden der Einstellungen
   * Änderungen an diesen vorzunehmen, bevor sie verwendet werden. Die ggf.
   * modifizierten Einstellungen werden von der Methode zurückgeliefert.
   *
   * @param array $settings Einstellungen des Addons
   * @return array
   */
  protected function _processAddonSettingsAfterLoad($settings)
  {
    return $settings;
  }

  /**
   * Modifiziert die Addon-Einstellungen bevor sie gespeichert werden.
   *
   * Diese Methode kann überladen werden, um vor dem Speichern der
   * Einstellungen Änderungen an diesen vorzunehmen. Die ggf. modifizierten
   * Einstellungen werden von der Methode zurückgeliefert.
   *
   * @param array $settings Einstellungen des Addons
   * @return array
   */
  protected function _processAddonSettingsBeforeSave($settings)
  {
    return $settings;
  }

  /**
   * Liefert die Zuordnungen von Auftragsdaten an benutzerdefinierte Felder.
   *
   * @param array $settings Array mit Einstellungen. Wird kein Array angegeben, so werden die Einstellungen vom
   *   GREYHOUND Server abgerufen.
   * @return array
   */
  public function getUserFieldVars($settings = null)
  {
    $userFieldVars = array();

    if(!is_array($settings))
      $settings = $this->getAddonSettings();

    $i = 0;

    while(isset($settings['UserFieldVar.' . $i . '.Var']))
    {
      $userFieldVars[] = array('var' => $settings['UserFieldVar.' . $i . '.Var'], 'field' => $settings['UserFieldVar.' . $i . '.Field']);
      $i++;
    }

    return $userFieldVars;
  }

  /**
   * Liefert die Felder für die Vorschauzeile.
   *
   * @param array $settings Array mit Einstellungen. Wird kein Array angegeben, so werden die Einstellungen vom
   *   GREYHOUND Server abgerufen.
   * @return array
   */
  public function getPreviewLineFields($settings = null)
  {
    $previewLineFields = array();

    if(!is_array($settings))
      $settings = $this->getAddonSettings();

    $i = 0;

    while(isset($settings['PreviewLineField.' . $i . '.Var']))
    {
      $previewLineFields[] = array('Var' => $settings['PreviewLineField.' . $i . '.Var'], 'Caption' => $settings['PreviewLineField.' . $i . '.Caption']);
      $i++;
    }

    return $previewLineFields;
  }

  /**
   * Liefert die konfigurierten Tracking-Links.
   *
   * @param array $settings Array mit Einstellungen. Wird kein Array angegeben, so werden die Einstellungen vom
   *   GREYHOUND Server abgerufen.
   * @return array
   */
  public function getTrackingLinks($settings = null)
  {
    $trackingLinks = array();

    if(!is_array($settings))
      $settings = $this->getAddonSettings();

    $i = 0;

    while(isset($settings['TrackingLink.' . $i . '.Url']))
    {
      $trackingLinks[] = array('Prefix' => $settings['TrackingLink.' . $i . '.Prefix'], 'Provider' => $settings['TrackingLink.' . $i . '.Provider'], 'Url' => $settings['TrackingLink.' . $i . '.Url']);
      $i++;
    }

    return $trackingLinks;
  }

  /**
   * Liefert den Versanddienstleister anhand einer Paketnummer.
   * Wird ein Versanddienstleister übergeben, so wird dieser zurückgeliefert sofern kein Versanddienstleister anhand
   * der Paketnummer ermittelt werden kann.
   *
   * @param string $trackingId       Paketnummer
   * @param string $trackingProvider Versanddienstleister (optional)
   * @return string
   */
  public function getTrackingProvider($trackingId, $trackingProvider = null)
  {
    foreach($this->getTrackingLinks() as $link)
      if(GhCoreStrings::startsWith($trackingId, $link['Prefix'], true))
        return $link['Provider'];

    return $trackingProvider;
  }

  /**
   * Liefert den Sendungsverfolgungs-Link anhand einer Paketnummer.
   * Wird ein Versanddienstleister übergeben, so wird dieser für die Bestimmung des Links herangezogen, wenn kein
   * Tracking-Link anhand der Paketnummer ermittelt werden kann.
   *
   * @param string $trackingId       Paketnummer
   * @param string $trackingProvider Versanddienstleister (optional)
   * @param string $zipCode          Postleitzahl
   * @return string
   */
  public function getTrackingUrl($trackingId, $trackingProvider = null, $zipCode = null)
  {
    foreach($this->getTrackingLinks() as $link) {
      if(GhCoreStrings::startsWith($trackingId, $link['Prefix'], true)) {
        return str_ireplace(array('[$Nr]', '[$Zip]'), array(rawurlencode($trackingId), rawurlencode($zipCode)), $link['Url']);
      }
    }

    return GhAddonData::getTrackingUrl($trackingId, $trackingProvider, $zipCode);
  }

  /**
   * Liefert die URL zu einer Zahlungs-Transaktion, sofern verfügbar.
   * 
   * @param string $transactionId Transaktions-ID
   * @param string $paymentMethod Zahlungsart
   */
  public function getPaymentTransactionUrl($transactionId, $paymentMethod)
  {
    if (strpos(strtolower($paymentMethod), 'paypal') !== false && $transactionId) {
      return 'https://www.paypal.com/myaccount/transactions/details/' . $transactionId;
    } else {
      return null;
    }
  }

  /**
   * Liefert eine Kunden-ID anhand eines benutzerdefinierten Felds.
   *
   * @param int   $itemId     Element-ID
   * @param int   $itemKind   Element-Typ
   * @param array $contactIds IDs der klassifizierten Kontakte
   * @return string
   */
  public function getCustomerIdFromUserField($itemId, $itemKind, $contactIds)
  {
    $customerId = null;
    $customerIdUserField = $this->getAddonSetting('CustomerIDUserField');

    if(!$customerIdUserField)
      return null;

    try
    {
      // Benutzerdefiniertes Feld am Element abrufen:
      $userFields = rpc_itemslib::newRpcItems($this->getGreyhoundClient())->GetUserFieldList($itemId, $itemKind);

      foreach($userFields as $userField)
      {
        /* @var GhRpcItemslibRpcUserFieldItem $userField */
        if($userField->FieldRef == $customerIdUserField)
        {
          $customerId = $userField->Value;
          break;
        }
      }

      // Falls keine Kunden-ID in einem benutzerdefinierten Feld am Element existiert, dann wird in benutzerdefinierten Feldern von klassifizierten Kontakten gesucht:
      if(!$customerId && !empty($contactIds))
      {
        $filter = rpc_itemslib::newRpcItemFilter();
        $filter->ItemRefs = $contactIds;

        foreach(rpc_itemslib::newRpcItems($this->getGreyhoundClient())->GetList($filter, rpc_itemslib::iobStartDate, rpc_itemslib::iodDescending, 0, 0, array($customerIdUserField), 0) as $contact)
        {
          if(is_array($contact->UserValues))
          {
            foreach($contact->UserValues as $userValue)
            {
              if($userValue->FieldRef = $customerIdUserField && $userValue->Value)
              {
                $customerId = $userValue->Value;
                break; // Nur die erste gefundene Kunden-ID verwenden
              }
            }
          }
        }
      }
    }
    catch(Exception $e)
    {
    }

    return trim($customerId);
  }

  /**
   * Liefert eine Kundennummer anhand eines benutzerdefinierten Felds.
   *
   * @param int   $itemId     Element-ID
   * @param int   $itemKind   Element-Typ
   * @param array $contactIds IDs der klassifizierten Kontakte
   * @return string
   */
  public function getCustomerNumberFromUserField($itemId, $itemKind, $contactIds)
  {
    $customerNo = null;
    $customerNoUserField = $this->getAddonSetting('CustomerNoUserField');

    if(!$customerNoUserField)
      return null;

    try
    {
      // Benutzerdefiniertes Feld am Element abrufen:
      $userFields = rpc_itemslib::newRpcItems($this->getGreyhoundClient())->GetUserFieldList($itemId, $itemKind);

      foreach($userFields as $userField)
      {
        /* @var GhRpcItemslibRpcUserFieldItem $userField */
        if($userField->FieldRef == $customerNoUserField)
        {
          $customerNo = $userField->Value;
          break;
        }
      }

      // Falls keine Kunden-ID in einem benutzerdefinierten Feld am Element existiert, dann wird in benutzerdefinierten Feldern von klassifizierten Kontakten gesucht:
      if(!$customerNo && !empty($contactIds))
      {
        $filter = rpc_itemslib::newRpcItemFilter();
        $filter->ItemRefs = $contactIds;

        foreach(rpc_itemslib::newRpcItems($this->getGreyhoundClient())->GetList($filter, rpc_itemslib::iobStartDate, rpc_itemslib::iodDescending, 0, 0, array($customerNoUserField), 0) as $contact)
        {
          if(is_array($contact->UserValues))
          {
            foreach($contact->UserValues as $userValue)
            {
              if($userValue->FieldRef = $customerNoUserField && $userValue->Value)
              {
                $customerNo = $userValue->Value;
                break; // Nur die erste gefundene Kunden-ID verwenden
              }
            }
          }
        }
      }
    }
    catch(Exception $e)
    {
    }

    return trim($customerNo);
  }

  /**
   * Liefert die E-Mail Adressen von Kontakten.
   *
   * @param array $contactIds Element-IDs der Kontakte
   * @return array
   */
  public function getEmailsFromContacts($contactIds)
  {
    if(!is_array($contactIds) || empty($contactIds))
      return null;

    $emails = array();
    $filter = rpc_itemslib::newRpcItemFilter();
    $filter->ItemRefs = $contactIds;

    foreach(rpc_itemslib::newRpcItems($this->getGreyhoundClient())->GetList($filter, rpc_itemslib::iobStartDate, rpc_itemslib::iodDescending, 0, GhRpcUtils::newSet(rpc_itemslib::cvContactEmail), array(), 0) as $contact)
    {
      $email = trim(GhAddonData::safeGet($contact, 'ContactEmail'), '<> ');

      if($email && !in_array($email, $emails))
        $emails[] = $email;
    }

    return $emails;
  }

  /**
   * Liefert die Template-Variablen und ihre Bezeichnung / Kurzbeschreibung.
   *
   * @return array
   */
  public function getTemplateVariableDescriptions()
  {
    return array(
      //'Order.OrderNumber' => 'Bestellnummer',
      //'Order.InvoiceNumber' => 'Rechnungsnummer',
      //'Order.InvoiceDate' => 'Rechnungsdatum',
      //'Order.CustomerNumber' => 'Kundennummer',
      //'Order.Date' => 'Bestelldatum',
      //'Order.Status' => 'Status des Auftrags',
      //'Order.Platform' => 'Verkaufsplattform',
      //'Order.PlatformUsername' => 'Benutzer der Verkaufsplattform',
      //'Order.TotalAmount' => 'Gesamtbetrag des Auftrags',
      //'Order.OpenAmount'=> 'Offener Betrag des Auftrags',
      //'BillingAddress.Company' => 'Rechnungsadresse: Firma',
      //'BillingAddress.Salutation' => 'Rechnungsadresse: Anrede',
      //'BillingAddress.Sex.Number' => 'Rechnungsadresse: Geschlecht',
      //'BillingAddress.Title' => 'Rechnungsadresse: Titel',
      //'BillingAddress.Name' => 'Rechnungsadresse: Name',
      //'BillingAddress.Firstname' => 'Rechnungsadresse: Vorname',
      //'BillingAddress.Lastname' => 'Rechnungsadresse: Nachname',
      //'BillingAddress.Letter.Identifier' => 'Rechnungsadresse: Adresse',
      //'BillingAddress.Letter.Street' => 'Rechnungsadresse: Straße',
      //'BillingAddress.Letter.AdditionalInfo' => 'Rechnungsadresse: Zusatzinfo',
      //'BillingAddress.Letter.ZipCode' => 'Rechnungsadresse: Postleitzahl',
      //'BillingAddress.Letter.City' => 'Rechnungsadresse: Stadt',
      //'BillingAddress.Letter.Province' => 'Rechnungsadresse: Bundesstaat/Provinz',
      //'BillingAddress.Letter.Country' => 'Rechnungsadresse: Land',
      //'BillingAddress.Letter.Country.ISO2' => 'Rechnungsadresse: Ländercode',
      //'BillingAddress.Email' => 'Rechnungsadresse: E-Mail Adresse',
      //'BillingAddress.Phone' => 'Rechnungsadresse: Telefonnummer',
      //'BillingAddress.Mobile' => 'Rechnungsadresse: Mobiltelefonnummer',
      //'BillingAddress.Fax' => 'Rechnungsadresse: Faxnummer',
      //'BillingAddress.Birthday' => 'Rechnungsadresse: Geburtstag',
      //'ShippingAddress.Company' => 'Lieferadresse: Firma',
      //'ShippingAddress.Salutation' => 'Lieferadresse: Anrede',
      //'ShippingAddress.Sex.Number' => 'Lieferadresse: Geschlecht',
      //'ShippingAddress.Title' => 'Lieferadresse: Titel',
      //'ShippingAddress.Name' => 'Lieferadresse: Name',
      //'ShippingAddress.Firstname' => 'Lieferadresse: Vorname',
      //'ShippingAddress.Lastname' => 'Lieferadresse: Nachname',
      //'ShippingAddress.Letter.Identifier' => 'Lieferadresse: Adresse',
      //'ShippingAddress.Letter.Street' => 'Lieferadresse: Straße',
      //'ShippingAddress.Letter.AdditionalInfo' => 'Lieferadresse: Zusatzinfo',
      //'ShippingAddress.Letter.ZipCode' => 'Lieferadresse: Postleitzahl',
      //'ShippingAddress.Letter.City' => 'Lieferadresse: Stadt',
      //'ShippingAddress.Letter.Province' => 'Lieferadresse: Bundesstaat/Provinz',
      //'ShippingAddress.Letter.Country' => 'Lieferadresse: Land',
      //'ShippingAddress.Letter.Country.ISO2' => 'Lieferadresse: Ländercode',
      //'ShippingAddress.Email' => 'Lieferadresse: E-Mail Adresse',
      //'ShippingAddress.Phone' => 'Lieferadresse: Telefonnummer',
      //'ShippingAddress.Mobile' => 'Lieferadresse: Mobiltelefonnummer',
      //'ShippingAddress.Fax' => 'Lieferadresse: Faxnummer',
      //'ShippingAddress.Birthday' => 'Lieferadresse: Geburtstag',
      //'Payment.Method' => 'Zahlungsart',
      //'Payment.Costs' => 'Kosten der Zahlungsart',
      //'Payment.Date' => 'Zahlungsdatum',
      //'Payment.TransactionID' => 'Transaktions-ID der Zahlung',
      //'Payment.TermOfPayment' => 'Zahlungsziel',
      //'Shipping.Provider' => 'Versanddienstleister',
      //'Shipping.Costs' => 'Versandkosten',
      //'Shipping.Date' => 'Lieferdatum',
      //'Shipping.Tracking.Code' => 'Sendungsnummer',
      //'Shipping.Tracking.Codes' => 'Alle Sendungsnummern',
      //'Shipping.Tracking.URL' => 'Sendungsverfolgungs-URL',
      //'Shipping.Tracking.URLs' => 'Alle Sendungsverfolgungs-URLs',
      //'Shipping.Tracking.Count' => 'Anzahl der Sendungsnummern',
      //'Shipping.Tracking.Code1' => 'Erste Sendungsnummer',
      //'Shipping.Tracking.URL1' => 'Erste Sendungsverfolgungs-URL',
      //'Shipping.Tracking.Code2' => 'Zweite Sendungsnummer',
      //'Shipping.Tracking.URL2' => 'Zweite Sendungsverfolgungs-URL',
      //'Shipping.Tracking.Code3' => 'Dritte Sendungsnummer',
      //'Shipping.Tracking.URL3' => 'Dritte Sendungsverfolgungs-URL'
    );
  }

  /**
   * Liefert eine Liste der Template-Variablen, die für die Suche verwendet werden können.
   *
   * @return array
   */
  public function getSearchTemplateVariables()
  {
    return array('Order.OrderNumber', 'Order.CustomerNumber');
  }

  /**
   * Liefert die Voreinstellungen für einen neuen Account.
   *
   * @return array
   */
  public function getNewAccountSettings()
  {
    return array();
  }

  /**
   * Liefert den Namen/Identifier des benutzerdefinierten Felds für die Kunden-ID des Drittanbieters.
   *
   * @return string
   */
  public function getCustomerIdUserFieldName()
  {
    return $this->_customerIdUserFieldName;
  }

  /**
   * Liefert die Beschreibung des benutzerdefinierten Felds für die Kunden-ID des Drittanbieters.
   *
   * @return string
   */
  public function getCustomerIdUserFieldDescription()
  {
    return $this->_customerIdUserFieldDescription;
  }

  /**
   * Erzeugt ein benutzerdefiniertes Feld für die Kunden-ID des Drittanbieters und liefert die ID des neuen
   * benutzerdefinierten Felds.
   *
   * @return int|string
   */
  public function createCustomerIdUserField()
  {
    $userField = rpc_userfieldslib::newRpcUserField();
    $userField->Name = $this->_customerIdUserFieldName;
    $userField->Description = $this->_customerIdUserFieldDescription;
    $userField->DataType = $this->_customerIdUserFieldType;
    $userField->Kinds = GhRpcUtils::newSet(rpc_userfieldslib::ufkContact);

    return rpc_userfieldslib::newRpcUserFields($this->getGreyhoundClient())->New_($userField);
  }

  /**
   * Liefert eine URL, die einen Link im Standard-Webbrowser öffnet.
   * Da die Erweiterungen im GREYHOUND Client und nicht im Webbrowser ausgeführt werden, müssen Browser-Links über einen
   * speziellen Link aufgerufen werden. Dieser Link sorgt dafür, dass eine GREYHOUND Erweiterung die gewünschte URL im
   * Standard-Webbrowser des Betriebssystems öffnet.
   * 
   * @param string $url URL, die im Webbrowser aufgerufen werden soll
   * @return string
   */
  public function getBrowserLink($url)
  {
    if ($url === null || $url === false || !strlen($url)) {
      return null;
    } elseif (GhCoreStrings::startsWith($url, 'greyhound:')) {
      return $url;
    }
    
    $extension = $this->getWebIdentifier();

    if ($extension) {
      return 'greyhound:ExecExtension?Identifier=' . $extension . '&url=' . rawurlencode(html_entity_decode($url));
    } else {
      return null;
    }
  }

  /**
   * Liefert den Namen der Debug-Log Datei.
   *
   * @return string
   */
  public function getDebugLogFile()
  {
    return null;
  }

  /**
   * Schreibt eine Nachricht in das Debug-Log, sofern der Debug-Modus aktiviert ist (Addon-Einstellung "DebugMode").
   *
   * @param string $message Nachricht
   * @param boolean $noHeader true: keinen Header (Datum, Uhrzeit) voranstellen, false: Datum und Uhrzeit voranstellen
   * @param boolean $noLineBreak true: keine Zeilenumbrüche hinter der Nachricht einfügen, false: zwei Zeilenumbrüche anfügen
   */
  public function log($message, $noHeader = false, $noLineBreak = false)
  {
    if($this->getAddonSetting('DebugMode', false) && $this->getDebugLogFile())
    {
      file_put_contents($this->getDebugLogFile(), ($noHeader ? '' : '[' . date('Y-m-d H:i:s') . ']  ')
        . $message
        . ($noLineBreak ? '' : "\n\n"),
        FILE_APPEND);
    }
  }
}