<?php
/*
 * GREYHOUND Addon Framework
 *
 * Copyright (c) 2011-2019 GREYHOUND Software GmbH & Co. KG
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of the GREYHOUND Addon Framework and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including without
 * limitation the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to permit persons to
 * whom the Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/**
 * Account eines Drittsystems zum Abruf von Daten.
 *
 * @package GhAddonFramework
 * @subpackage core
 * @category PHP Addon Framework
 *
 * @author digital guru GmbH &amp; Co. KG <develop@greyhound-software.com>
 * @copyright 2011-2019 digital guru GmbH &amp; Co. KG
 * @link greyhound-software.com
 */
abstract class GhAddonAccount
{
  /**
   * Addon.
   * @var GhAddon
   */
  protected $_addon;

  /**
   * Account ID.
   * @var int
   */
  protected $_id;

  /**
   * Account-Einstellungen.
   * @var array
   */
  protected $_settings;

  /**
   * Empfängerfilter.
   * @var array
   */
  protected $_recipientFilters;

  /**
   * Templates (GREYHOUND Textbausteine).
   * @var array
   */
  protected $_templates;

  /**
   * Cache-Timeout (in Sekunden) für Templates.
   * @var int
   */
  protected $_templatesCacheTime = 300; // 5 Minuten


  /**
   * Konstruktor.
   *
   * @param GhAddon $addon    Addon
   * @param int     $id       Account-ID
   * @param array   $settings Account-Einstellungen
   */
  public function __construct($addon, $id = null, $settings = null)
  {
    $this->_addon = $addon;
    $this->_id = $id;
    $this->_settings = $settings ? $settings : array();
  }

  /**
   * Diese Methode muss überladen werden und true liefern falls der Account
   * gültig ist (z.B. über alle notwendigen Einstellungen für die Zugangsdaten
   * zum Drittsystem verfügt).
   *
   * @return boolean
   */
  abstract public function isValid();

  /**
   * Prüft, ob der Account aktiv oder inaktiv ist.
   *
   * @return boolean
   */
  public function isActive()
  {
    if ($this->getSetting('Activated')) {
      return true;
    } else {
      return false;
    }
  }

  /**
   * Erzeugt ein Datenobjekt, das Aufträge und Kunden repräsentiert.
   *
   * @param array $data Auftrags- und Kundendaten (Addon-spezifisch)
   * @return GhAddonData
   */
  abstract public function newOrderData($data);

  /**
   * Die Methode sorgt dafür, dass die Templates nicht im Account serialisiert
   * werden.
   *
   * @return array
   */
  public function __sleep()
  {
    $vars = array_keys(get_object_vars($this));
    unset($vars['_templates']);
    unset($vars['_templatesCacheTime']);

    return $vars;
  }

  /**
   * Leert den Cache dieses Accounts.
   */
  public function clearCache()
  {
    $this->_clearTemplatesCache();
  }

  /**
   * Liefert die ID des Accounts.
   *
   * @return int
   */
  public function getId()
  {
    return $this->_id;
  }

  /**
   * Liefert das Addon in dem dieser Account verwendet wird.
   *
   * @return GhAddon
   */
  public function getAddon()
  {
    return $this->_addon;
  }

  /**
   * Liefert die Einstellungen des Accounts.
   *
   * @return array
   */
  public function getSettings()
  {
    return $this->_settings;
  }

  /**
   * Liefert eine Account-Einstellung.
   *
   * @param string $setting Name der Einstellung
   * @param string $default Vorgabewert falls die Einstellung nicht gesetzt ist
   * @return string
   */
  public function getSetting($setting, $default = null)
  {
    $settings = $this->getSettings();

    if(isset($settings[$setting]))
      return $settings[$setting];
    else
      return $default;
  }

  /**
   * Liefert den Empfänger-Filter des Accounts als regulären Ausdruck.
   *
   * @return array
   */
  public function getRecipientFilters()
  {
    if(!is_array($this->_recipientFilters))
    {
      $this->_recipientFilters = array();
      $i = 0;
      $recipientFilter = GhAddonData::safeGet($this->_settings, 'Recipient' . ($i++));

      while($recipientFilter)
      {
        $recipientFilter = trim($recipientFilter);

        if(strlen($recipientFilter) > 0)
          $this->_recipientFilters[] = $recipientFilter;

        $recipientFilter = GhAddonData::safeGet($this->_settings, 'Recipient' . ($i++));
      }

      // Falls keine Filter definiert wurden, dann wird '*' angenommen:
      if(empty($this->_recipientFilters))
        $this->_recipientFilters[] = '*';

      // Filter in reguläre Ausdrücke umwandeln:
      if(!empty($this->_recipientFilters))
      {
        foreach($this->_recipientFilters as $key => $recipientFilter)
        {
          $recipientFilter = str_replace('.', '\\.', $recipientFilter);
          $recipientFilter = str_replace(array('?', '*', '+', '(', ')', '[', ']'), array('.', '.*', '\\+', '\\(', '\\)', '\\[', '\\]'), $recipientFilter);

          $this->_recipientFilters[$key] = '/^' . $recipientFilter . '$/';
        }
      }
    }

    return $this->_recipientFilters;
  }

  /**
   * Prüft ob für einen Empfänger Daten abgerufen werden sollen.
   *
   * Falls Daten für den Empfänger abgerufen werden (gemäß der Filter-Einstellungen
   * des Addons), so wird true zurückgeliefert. Falls keine Daten abgerufen
   * werden brauchen wird false geliefert.
   *
   * @param string|array $recipients Empfänger (oder Array mit Empfängern)
   * @return boolean
   */
  public function isValidRecipient($recipients)
  {
    if(empty($recipients))
      return true;

    if(!is_array($recipients))
      $recipients = array($recipients);

    $filters = $this->getRecipientFilters();

    if(empty($filters))
      return false;

    foreach($recipients as $recipient)
      foreach($filters as $filter)
        if(preg_match($filter, $recipient))
          return true;

    return false;
  }

  /**
   * Liefert eine Liste der Templates (GREYHOUND Textbausteine).
   * Die Liste wird im Session-Cache vorgehalten, da sich die Templates nur
   * sehr selten ändern werden.
   *
   * @param boolean $refreshCache true: Daten nicht aus dem Cache verwenden, Cache ggf. erneuern, false: Daten aus dem Cache verwenden, falls vorhanden
   * @return array
   */
  public function getTemplates($refreshCache = false)
  {
    if($refreshCache)
      $this->_clearTemplatesCache();

    if(!is_array($this->_templates))
    {
      $templatesStr = GhCoreGlobals::session('templates');

      if($templatesStr)
      {
        $this->_templates = unserialize($templatesStr);

        if(!is_array($this->_templates))
          $this->_templates = array();
      }
    }

    if(!isset($this->_templates[$this->getId()]))
    {
      $this->_templates[$this->getId()] = array();
      $textBlockRootId = $this->_addon->getAddonSetting('TextBlockRoot');

      if($textBlockRootId)
      {
        // Textbausteine abrufen und den ausgewählten Teilbaum als Vorlagen verwenden:
        $textBlocks = rpc_textblockslib::newRpcTextBlocks($this->_addon->getGreyhoundClient())->GetList(true);

        if($textBlockRootId != -1)
          $textBlocksTree = new GhRpcTree($textBlocks, 'Name', 'ParentRef', $textBlockRootId);
        else
          $textBlocksTree = new GhRpcTree($textBlocks, 'Name', 'ParentRef');

        $textBlocks = $textBlocksTree->getList();

        foreach($textBlocks as $treeNode)
        {
          /* @var GhCoreTreeNode $treeNode */
          $textBlock = $treeNode->getData();/* @var GhRpcTextblockslibRpcTextBlock $textBlock */

          // Textbausteine, die auf der Hauptebene liegen sollen, dürfen keine Parent-ID haben, sonst gibt es Fehler im GhCoreTree:
          if($textBlock->ParentRef == $textBlockRootId)
            $textBlock->ParentRef = 0;

          $template = new GhAddonTemplate();
          $template->ID = $textBlock->ID;
          $template->ParentRef = $textBlock->ParentRef;
          $template->Name = GhCoreStrings::attributeValue($textBlock->Name);
          $template->Subject = GhCoreStrings::attributeValue($textBlock->Subject);
          $template->IsHtml = $textBlock->HtmlData ? true : false;

          if($textBlock->HtmlData)
            $template->Description = trim($textBlock->HtmlData);
          else // nl2br() scheint hier teilweise Probleme zu machen, daher werden die Zeilenumbrüche durch str_replace ersetzt:
            $template->Description = str_replace("\r\n", '<br />', trim($textBlock->TextData));

          // Alle script Tags müssen entfernt werden, da Javascript-Strings bei </script> abbrechen, auch wenn das </script> innerhalb eines Strings auftritt:
          $template->Description = preg_replace('/<script[\s>].*<\/script>/i', '', $template->Description);

          $template->Description = iconv('windows-1252', 'UTF-8', $template->Description);

          $this->_templates[$this->getId()][$template->ID] = $template;
        }
      }

      GhCoreGlobals::session('templates', serialize($this->_templates), true);
    }

    return $this->_templates[$this->getId()];
  }

  /**
   * Leert den Template-Cache.
   */
  protected function _clearTemplatesCache()
  {
    $this->_templates = false;
    GhCoreGlobals::session('templates', null, true);
    GhCoreGlobals::session('templatesTime', null, true);
  }

  /**
   * Liefert den Inhalt einer E-Mail Vorlage.
   *
   * Der Rückgabewert ist ein Array mit den Indices 'Subject', 'Content' und
   * 'Format' für den Betreff bzw. den Inhalt und das Inhaltsformat (html/text)
   * der Vorlage.
   *
   * @param GhAddonTemplate $template Vorlage
   * @param GhAddonOrder    $order    Bestellung
   * @param GhAddonCustomer $customer Kunde
   * @param string          $format   Format für den Template-Inhalt (html, text, auto)
   * @return string
   */
  protected function _getTemplateContent($template, $order, $customer, $format = null)
  {
    if(!$template)
      return null;

    $textBlock = rpc_textblockslib::newRpcTextBlocks($this->_addon->getGreyhoundClient())->Get($template->ID, false);

    if(!$textBlock)
      return null;

    if($format == GhAddonData::FORMAT_HTML && !$textBlock->HtmlData && $textBlock->TextData)
      $format = GhAddonData::FORMAT_TEXT;
    elseif($format == GhAddonData::FORMAT_TEXT && !$textBlock->TextData && $textBlock->HtmlData)
      $format = GhAddonData::FORMAT_HTML;
    else
      $format = $template->IsHtml ? GhAddonData::FORMAT_HTML : GhAddonData::FORMAT_TEXT;

    $data = array(
      'Subject' => $textBlock->Subject,
      'Content' => $format == GhAddonData::FORMAT_HTML ? $textBlock->HtmlData : $textBlock->TextData,
      'Format' => $format
    );

    $data['Content'] = iconv('windows-1252', 'UTF-8', $data['Content']);

    return $data;
  }

  /**
   * Sucht nach Kunden und Bestellungen.
   *
   * Die Suche wird durch Parameter eingegrenzt, die als Array übergeben
   * werden. Dabei sind folgende Array-Keys möglich:
   * search     : string, Suchbegriff (sucht beim Drittanbieter, falls keine
   *              Treffer erzielt werden wird nach dem Suchebgriff in den
   *              GREYHOUND Kontakten gesucht und dann beim Drittanbieter nach
   *              der Kunden-ID der gefundenen GREYHOUND Kontakte)
   * email      : string, sucht beim Drittanbieter nach dieser E-Mail-Adresse
   * ebay       : string, sucht beim Drittanbieter nach diesem eBay-Namen
   * taskNumber : string, Vorgangsnummer des ausgewählten Elements
   * action     : string, Aktion (selectCustomer, selectOrder), um bestimmte
   *              Suchparameter zu bevorzugen
   *
   * @throws Exception falls beim Abrufen der Daten ein Fehler auftritt
   *
   * @param int   $customerId Kunden-ID
   * @param int   $orderId    Bestell-ID
   * @param array $params     Suchparameter
   * @return GhAddonData
   */
  public function search($customerId, $orderId, $params)
  {
    $searchTerm = isset($params['search']) ? $params['search'] : null;
    $email = isset($params['email']) ? (string)$params['email'] : null;
    $ebayName = isset($params['ebay']) ? (string)$params['ebay'] : null;
    $amazonId = isset($params['amazon']) ? (string)$params['amazon'] : null;
    $itemTaskNumber = isset($params['taskNumber']) ? (string)$params['taskNumber'] : null;
    $action = isset($params['action']) ? $params['action'] : null;

    $data = null;

    // Nach Suchbegriff suchen:
    if($searchTerm)
    {
      $data = $this->searchBySearchTerm($searchTerm, $params);

      // Falls beim Drittanbieter nichts gefunden wurde wird im GREYHOUND nach Kontakten mit einer Kunden-ID des Drittanbieters gesucht:
      if(empty($data))
      {
        try
        {
          $customerIds = $data = $this->searchGreyhoundCustomerIds($searchTerm, $params);

          if(!empty($customerIds))
            $data = $this->searchByCustomerIds($customerIds, $params);
        }
        catch(Exception $e)
        {
        }
      }
    }
    elseif($action == 'selectUser' && $customerId)
      $data = $this->searchByCustomerIds(array($customerId), $params);
    elseif($action == 'selectOrder' && $orderId)
      $data = $this->_searchSelectOrder($customerId, $orderId);
    else
    {
      // Bestelldaten anhand der Vorgangsnummer ermitteln:
      if(!$orderId && $itemTaskNumber && $itemTaskNumber[0] == $this->_addon->getAddonSetting('TaskNumberPrefix'))
        $data = $this->searchByTaskNumber($itemTaskNumber, $params);

      // Nach Kunden-ID suchen:
      if(!$data && $customerId)
        $data = $this->searchByCustomerIds(array($customerId), $params);

      if(!$data && $amazonId)
        $data = $this->searchByAmazonId($amazonId, $params);

      // Nach eBay-Namen suchen:
      if(!$data && $ebayName) {
        $data = $this->searchByEbayName($ebayName, $params);

        if($data && !$data->getSelectedOrderId()) {
          foreach($data->getOrders() as $order) {
            if($order && $order->PlatformUsername && strtolower($order->PlatformUsername) == strtolower($ebayName)) {
              $data->setSelectedOrderId($order->ID);
              break;
            }
          }
        }
      }

      // Nach E-Mail Adresse suchen:
      if(!$data && $email)
        $data = $this->searchByEmail($email, $params);

      if(!$data)
        $data = false;
    }

    return $data;
  }

  /**
   * Sucht im GREYHOUND nach Kunden-IDs des Drittanbieters.
   * Hierzu wird nach Kontakten gesucht, die ein benutzerdefiniertes Feld mit
   * einer Kunden-ID haben. Es wird ein Array mit den gefundenen Kunden-IDs des
   * Drittanbieters zurück geliefert.
   *
   * @param string $searchTerm Suchbegriff
   * @param array $params Alle Suchparameter
   * @return array
   */
  public function searchGreyhoundCustomerIds($searchTerm, $params = null)
  {
    $customerIds = array();
    $customerIdUserField = $this->_addon->getAddonSetting('CustomerIDUserField');

    if($customerIdUserField)
    {
      $filter = rpc_itemslib::newRpcItemFilter();
      $filter->Kinds = GhRpcUtils::newSet(rpc_itemslib::ikContact);
      $filter->NegativeFlags = GhRpcUtils::newSet(rpc_itemslib::ifDeleted, rpc_itemslib::ifSpam);
      $filter->SearchFor = $searchTerm;
      $filter->SearchFields = GhRpcUtils::newSet(rpc_itemslib::isfAddress, rpc_itemslib::isfSubject, rpc_itemslib::isfData);

      $itemService = rpc_itemslib::newRpcItems($this->_addon->getGreyhoundClient());

      try
      {
        $contacts = $itemService->GetList($filter, rpc_itemslib::iobStartDate, rpc_itemslib::iodDescending, 0, 0, array($customerIdUserField), 0);

        foreach($contacts as $tmpContact)
        {
          if(!is_array($tmpContact->UserValues))
            continue;

          $tmpUserId = null;

          foreach($tmpContact->UserValues as $tmpUserValue)
          {
            if($tmpUserValue->FieldRef == $customerIdUserField)
            {
              if(!in_array($tmpUserValue->Value, $customerIds))
              {
                $customerIds[] = $tmpUserValue->Value;
                break;
              }
            }
          }
        }
      }
      catch(Exception $e)
      {
      }
    }

    return $customerIds;
  }

  /**
   * Sucht nach einer ausgewählten Bestellung und liefert die Bestellungen des
   * Kunden zurück, wobei die gewählte Bestellung vorausgewählt wird.
   *
   * @param string $customerId Kunden-ID
   * @param string $orderId    Bestell-ID
   * @return GhAddonData
   */
  protected function _searchSelectOrder($customerId, $orderId)
  {
    if ($orderId) {
      $data = $this->searchByOrderIds(array($orderId));
    } elseif ($customerId) {
      $data = $this->searchByCustomerIds(array($customerId));
    } else {
      $data = null;
    }

    if ($data) {
      $data->setSelectedOrderId($orderId);
    }

    return $data;
  }

  /**
   * Diese Methode kann überladen werden, um beim Drittanbieter nach einem
   * Suchbegriff zu suchen und Kunden und Bestellungen zu liefern.
   *
   * @param string $searchTerm Suchbegriff
   * @param array $params Alle Suchparameter
   * @return GhAddonData
   */
  public function searchBySearchTerm($searchTerm, $params = null)
  {
    return null;
  }

  /**
   * Diese Methode kann überladen werden, um beim Drittanbieter nach einer
   * E-Mail Adresse zu suchen und Kunden und Bestellungen zu liefern.
   *
   * @param string $email E-Mail Adresse
   * @param array $params Alle Suchparameter
   * @return GhAddonData
   */
  public function searchByEmail($email, $params = null)
  {
    return null;
  }

  /**
   * Diese Methode kann überladen werden, um beim Drittanbieter nach einer
   * Amazon Order-ID zu suchen und Kunden und Bestellungen zu liefern.
   *
   * @param string $amazonId Amazon Order-ID
   * @param array $params Alle Suchparameter
   * @return GhAddonData
   */
  public function searchByAmazonId($amazonId, $params = null)
  {
    return null;
  }

  /**
   * Diese Methode kann überladen werden, um beim Drittanbieter nach einem
   * eBay Namen zu suchen und Kunden und Bestellungen zu liefern.
   *
   * @param string $ebayName eBay Name
   * @param array $params Alle Suchparameter
   * @return GhAddonData
   */
  public function searchByEbayName($ebayName, $params = null)
  {
    return null;
  }

  /**
   * Diese Methode kann überladen werden, um beim Drittanbieter nach Kunden-IDs
   * zu suchen und Kunden und Bestellungen zu liefern.
   *
   * @param array $customerIds Kunden-IDs
   * @param array $params Alle Suchparameter
   * @return GhAddonData
   */
  public function searchByCustomerIds($customerIds, $params = null)
  {
    return null;
  }

  /**
   * Diese Methode kann überladen werden, um beim Drittanbieter nach Kundennummern
   * zu suchen und Kunden und Bestellungen zu liefern.
   *
   * @param array $customerNumbers Kundennummern
   * @param array $params Alle Suchparameter
   * @return GhAddonData
   */
  public function searchByCustomerNumbers($customerNumbers, $params = null)
  {
    return $this->searchByCustomerIds($customerNumbers);
  }

  /**
   * Diese Methode kann überladen werden, um beim Drittanbieter nach Bestell-IDs
   * zu suchen und Kunden und Bestellungen zu liefern.
   *
   * @param array $orderIds Bestell-IDs
   * @param array $params Alle Suchparameter
   * @return GhAddonData
   */
  public function searchByOrderIds($orderIds, $params = null)
  {
    return null;
  }

  /**
   * Sucht anhand der Vorgangsnummer nach Bestelldaten.
   *
   * @param string $taskNumber Vorgangsnummer
   * @param array $params Alle Suchparameter
   * @return GhAddonData
   */
  public function searchByTaskNumber($taskNumber, $params = null)
  {
    $data = null;

    // OrderID anhand der Vorgangsnummer ermitteln:
    if($taskNumber)
    {
      $orderId = $this->taskNumberToOrderId($taskNumber);

      if($orderId)
      {
        $data = $this->searchByOrderIds(array($orderId), $params);

        if($data)
        {
          $data->setSelectedOrderId($orderId);
          $tmpCustomers = $data->getCustomers();

          if(is_array($tmpCustomers) && !empty($tmpCustomers))
          {
            foreach($tmpCustomers as $tmpCustomer)
            {
              $customerId = $tmpCustomer->ID;
              $data = $this->searchByCustomerIds(array($customerId));

              if($data)
                $data->setSelectedOrderId($orderId);

              break; // Nur den ersten Eintrag betrachten, der allerdings nicht zwangsläufig den Index '0' hat (daher foreach)
            }
          }
        }
      }
    }

    return $data;
  }

  /**
   * Liefert eine E-Mail Vorlage des Drittanbieters.
   *
   * @param string $templateId ID des Templates
   * @return GhAddonTemplate
   */
  public function getTemplate($templateId)
  {
    $templates = $this->getTemplates();

    if(!isset($templates[$templateId]))
      return null;
    else
      return $templates[$templateId];
  }

  /**
   * Wendet eine E-Mail Vorlage auf eine Bestellung und einen Kunden an.
   * Diese Methode kann überladen werden, um einen Text für das Beantworten
   * oder Verfassen von E-Mails, Briefen, usw. bereitzustellen. Die Methode
   * muss dann ein Objekt mit den Eigenschaften 'Subject' und 'Html' liefern,
   * in denen der Betreff bzw. der HTML-Inhalt der Vorlage für die angegebene
   * Bestellung und den angegebenen Kunden enthalten ist.
   *
   * Der Rückgabewert ist ein Array mit den Indices 'Subject', 'Content' und
   * 'Format' für den Betreff bzw. den Inhalt und das Inhaltsformat (html/text)
   * der Vorlage.
   *
   * @param GhAddonTemplate $template Template
   * @param GhAddonOrder    $order    Bestellung
   * @param GhAddonCustomer $customer Kunde
   * @param string          $format   Format für den Template-Inhalt (html, text, auto)
   * @return array
   */
  public function applyTemplate($template, $order, $customer, $format = null)
  {
    $result = $this->_getTemplateContent($template, $order, $customer, $format);

    if(isset($result['Format']) && isset($result['Content']) && $format)
    {
      if($format == GhAddonData::FORMAT_HTML && $result['Format'] == GhAddonData::FORMAT_TEXT)
      {
        $result['Content'] = str_replace("\n", "<br>\r\n", str_replace("\r\n", "\n", $result['Content']));
      }
      elseif($format == GhAddonData::FORMAT_TEXT && $result['Format'] == GhAddonData::FORMAT_HTML)
      {
        $result['Content'] = str_replace("\n", "\r\n", str_replace("\r\n", "\n", $result['Content']));
        $result['Content'] = preg_replace('/<br[^>]>[^\r][^\n]/', "\r\n", $result['Content']);
        $result['Content'] = preg_replace('/<br[^>]>/', '', $result['Content']);
        $result['Content'] = strip_tags($result['Content']);
        $result['Content'] = html_entity_decode($result['Content'], ENT_QUOTES);

        $lines = explode("\r\n", $result['Content']);

        foreach($lines as $key => $line)
          $lines[$key] = trim($line);

        $result['Content'] = implode("\r\n", $lines);
      }
    }

    $variables = $this->getTemplateVariables($customer, $order, $template);

    if(!empty($variables))
    {
      $replaceKeys = array_keys($variables);
      $replaceValues = array_values($variables);

      if(isset($result['Subject']))
        $result['Subject'] = str_replace($replaceKeys, $replaceValues, $result['Subject']);
    }

    return $result;
  }

  /**
   * Liefert Variablen (Schlüssel-Wert-Paare) die in einer Vorlage ersetzt werden.
   *
   * @param GhAddonCustomer $customer Kunde
   * @param GhAddonOrder    $order    Bestellung
   * @param GhAddonTemplate $template Vorlage
   * @return array
   */
  public function getTemplateVariables($customer, $order, $template)
  {
    $currency = GhAddonData::safeGet($order, 'Currency');

    switch($currency)
    {
      case '€':
        $currencyCode = 'EUR';
        break;
      case 'EUR':
        $currency = '€';
        $currencyCode = 'EUR';
        break;
      case '$':
        $currencyCode = 'USD';
        break;
      case 'USD':
        $currency = '$';
        $currencyCode = 'USD';
        break;
      default:
        $currencyCode = $currency;
    }

    $variables = array(
      'Addon.Name' => $this->_addon->getName(),
      'Addon.Version' => $this->_addon->getVersion() . ' build ' . $this->_addon->getBuild(),
      'Addon.VersionInfo.Number' => $this->_addon->getVersion(),
      'Addon.VersionInfo.Build' => $this->_addon->getBuild(),
      'Order.OrderNumber' => GhAddonData::safeGet($order, 'OrderNumber'),
      'Order.InvoiceNumber' => GhAddonData::safeGet($order, 'InvoiceNumber'),
      'Order.Status' => GhAddonData::safeGet($order, 'Status'),
      'Order.CustomerNumber' => GhAddonData::safeGet($order, 'CustomerNumber'),
      'Order.Platform' => GhAddonData::safeGet($order, 'Platform'),
      'Order.PlatformUsername' => GhAddonData::safeGet($order, 'PlatformUsername'),
      'Payment.Method' => GhAddonData::safeGet($order, 'PaymentMethod'),
      'Payment.TransactionID' => GhAddonData::safeGet($order, 'PaymentTransactionID'),
      'Payment.TermOfPayment' => GhAddonData::safeGet($order, 'TermOfPayment'),
      'Shipping.Provider' => GhAddonData::safeGet($order, 'ShippingProvider'),
      'Shipping.Tracking.Code' => '',
      'Shipping.Tracking.Codes' => '',
      'Shipping.Tracking.URL' => '',
      'Shipping.Tracking.URLs' => '',
      'Shipping.Tracking.Count' => count($order->Tracking)
    );

    $index = 1;
    $trackingCodes = array();
    $trackingUrls = array();

    foreach($order->Tracking as $tracking)
    {
      if(!$variables['Shipping.Tracking.Code'])
      {
        $variables['Shipping.Tracking.Code'] = $tracking['Code'];
        $variables['Shipping.Tracking.URL'] = $tracking['Url'];
      }

      $variables['Shipping.Tracking.Code' . $index] = $tracking['Code'];
      $variables['Shipping.Tracking.URL' . $index] = $tracking['Url'];
      $trackingCodes[] = $tracking['Code'];
      $trackingUrls[] = $tracking['Url'];
      $index++;
    }

    $variables['Shipping.Tracking.Codes'] = implode(', ', $trackingCodes);
    $variables['Shipping.Tracking.URLs'] = implode(" ", $trackingUrls);

    // Datumsangaben:
    $dateVars = array(
      'Order.Date' => $order->Date,
      'Order.InvoiceDate' => $order->InvoiceDate,
      'Payment.Date' => $order->PaymentDate,
      'Shipping.Date' => $order->ShippingDate
    );

    foreach($dateVars as $var => $date)
    {
      $dateInfo = $this->_getDateInfo($date);
      $variables[$var] = trim($dateInfo['D'] . '.' . $dateInfo['M'] . '.' . $dateInfo['Y'], '.');

      foreach($dateInfo as $key => $value)
        $variables[$var . 'Info.' . $key] = $value;
    }

    // Preise:
    $priceVars = array(
      'Order.TotalAmount' => GhAddonData::safeGet($order, 'TotalAmount', 'float', true),
      'Order.OpenAmount' => GhAddonData::safeGet($order, 'OpenAmount', 'float', true),
      'Payment.Costs' => GhAddonData::safeGet($order, 'PaymentCosts', 'float', true),
      'Shipping.Costs' => GhAddonData::safeGet($order, 'ShippingCosts', 'float', true)
    );

    foreach($priceVars as $var => $price)
    {
      if($price !== null && $price !== false)
        $variables[$var] = GhCoreLocale::formatCurrency($price, $currency);
      else
        $variables[$var] = '';

      $variables[$var . 'Info.Value'] = $price;
      $variables[$var . 'Info.Currency'] = $currencyCode;
    }

    // Rechnungs- und Lieferadresse:
    $addressVars = array(
      'BillingAddress' => $order->BillingAddress,
      'ShippingAddress' => $order->ShippingAddress
    );

    foreach($addressVars as $addressType => $address)
    {
      $address = GhCoreLocale::fixAddress($address);  //  Wenn ich die Adresse über AJAX reinschiebe, würde sie hier zerstört werden.

      foreach($address as $key => $value)
      {
        if($value instanceof GhRpcDateTime)
        {
          $dateInfo = $this->_getDateInfo($value);

          if($value->getTimestamp())
            $variables[$addressType . '.' . $key] = $value->getFormatted('d.m.Y');
          else
            $variables[$addressType . '.' . $key] = '';

          foreach($dateInfo as $dateKey => $dateValue)
            $variables[$addressType . '.' . $key . 'Info.' . $dateKey] = $dateValue;
        }
        elseif(GhCoreStrings::endsWith($key, '.de'))
          $variables[$addressType . '.' . substr($key, 0, strlen($key) - 2) . 'DE'] = $value;
        elseif(GhCoreStrings::endsWith($key, '.DE'))
          $variables[$addressType . '.' . $key . '_UC'] = $value;
        elseif(GhCoreStrings::endsWith($key, '.en'))
          $variables[$addressType . '.' . substr($key, 0, strlen($key) - 2) . 'EN'] = $value;
        elseif(GhCoreStrings::endsWith($key, '.EN'))
          $variables[$addressType . '.' . $key . '_UC'] = $value;
        elseif(GhCoreStrings::endsWith($key, '.es'))
          $variables[$addressType . '.' . substr($key, 0, strlen($key) - 2) . 'ES'] = $value;
        elseif(GhCoreStrings::endsWith($key, '.ES'))
          $variables[$addressType . '.' . $key . '_UC'] = $value;
        elseif(GhCoreStrings::endsWith($key, '.fr'))
          $variables[$addressType . '.' . substr($key, 0, strlen($key) - 2) . 'FR'] = $value;
        elseif(GhCoreStrings::endsWith($key, '.FR'))
          $variables[$addressType . '.' . $key . '_UC'] = $value;
        else
          $variables[$addressType . '.' . $key] = $value;
      }
    }

    return $variables;
  }

  /**
   * Liefert eine Vorgangsnummer zu einer Auftrags-ID.
   *
   * @param string $orderId Auftrags-ID
   * @return string
   */
  public function orderIdToTaskNumber($orderId)
  {
    $orderId = strtoupper(sprintf('%010s', $orderId));

    // Der innere Teil einer Vorgangsnummer muss genau 10 Stellen haben:
    if(strlen($orderId) > 10)
      $orderId = substr($orderId, strlen($orderId) - 10);

    $prefix = $this->_addon->getAddonSetting('TaskNumberPrefix');

    if($prefix)
      return sprintf('%s-%010s-%02d', $prefix, $orderId, GhAddonData::getItemNumberChecksum($orderId));
    else
      return null;
  }

  /**
   * Liefert die Auftrags-ID zu einer Vorgangsnummer.
   *
   * @param string $taskNumber Vorgangsnummer
   * @return string
   */
  public function taskNumberToOrderId($taskNumber)
  {
    return substr($taskNumber, 2, 10);
  }

  /**
   * Liefert Datumsinformationen für Variablen.
   *
   * @param GhRpcDateTime $date Datum
   * @return array
   */
  protected function _getDateInfo($date)
  {
    $result = array('D' => '', 'M' => '', 'Y' => '', 'Day.DE' => '', 'Month.DE' => '', 'Day.EN' => '', 'Month.EN' => '', 'Day.ES' => '', 'Month.ES' => '', 'Day.FR' => '', 'Month.FR' => '', 'Timestamp' => 0);

    if($date instanceof GhRpcDateTime)
    {
      $timestamp = $date->getTimestamp();
      $result['Timestamp'] = (int)$timestamp;

      if($timestamp)
      {
        $result['D'] = GhCoreLocale::date('d', $timestamp);
        $result['M'] = GhCoreLocale::date('m', $timestamp);
        $result['Y'] = GhCoreLocale::date('Y', $timestamp);
        $result['Day.DE'] = GhCoreLocale::date('l', $timestamp, 'de');
        $result['Month.DE'] = GhCoreLocale::date('F', $timestamp, 'de');
        $result['Day.EN'] = GhCoreLocale::date('l', $timestamp, 'en');
        $result['Month.EN'] = GhCoreLocale::date('F', $timestamp, 'en');
        $result['Day.ES'] = GhCoreLocale::date('l', $timestamp, 'es');
        $result['Month.ES'] = GhCoreLocale::date('F', $timestamp, 'es');
        $result['Day.FR'] = GhCoreLocale::date('l', $timestamp, 'fr');
        $result['Month.FR'] = GhCoreLocale::date('F', $timestamp, 'fr');
      }
    }

    return $result;
  }

  /**
   * Liefert eine URL, die einen Kunden im Webbrowser oder einer Drittanwendung öffnet.
   *
   * @param GhAddonOrder $order Auftrag
   * @return string
   */
  public function getCustomerUrl($order)
  {
    return null;
  }

  /**
   * Liefert einen onclick-Handler, der einen Kunden im Webbrowser oder einer Drittanwendung öffnet.
   *
   * @param GhAddonOrder $order Auftrag
   * @return string
   */
  public function getCustomerClickHandler($order)
  {
    return null;
  }

  /**
   * Liefert eine URL, die einen Auftrag im Webbrowser oder einer Drittanwendung öffnet.
   *
   * @param GhAddonOrder $order Auftrag
   * @return string
   */
  public function getOrderUrl($order)
  {
    return null;
  }

  /**
   * Liefert einen onclick-Handler, der einen Auftrag im Webbrowser oder einer Drittanwendung öffnet.
   *
   * @param GhAddonOrder $order Auftrag
   * @return string
   */
  public function getOrderClickHandler($order)
  {
    return null;
  }

  /**
   * Liefert eine URL, die eine Rechnung im Webbrowser oder einer Drittanwendung öffnet.
   *
   * @param GhAddonOrder $order Auftrag
   * @return string
   */
  public function getInvoiceUrl($order)
  {
    return null;
  }

  /**
   * Liefert einen onclick-Handler, der eine Rechnung im Webbrowser oder einer Drittanwendung öffnet.
   *
   * @param GhAddonOrder $order Auftrag
   * @return string
   */
  public function getInvoiceClickHandler($order)
  {
    return null;
  }

  /**
   * Liefert eine URL, die einen Artikel im Webbrowser oder einer Drittanwendung öffnet.
   *
   * @param GhAddonOrderItem $orderItem Bestellposition
   * @param GhAddonOrder $order Auftrag
   * @return string
   */
  public function getProductUrl($orderItem, $order)
  {
    return null;
  }

  /**
   * Liefert einen onclick-Handler, der einen Artikel im Webbrowser oder einer Drittanwendung öffnet.
   *
   * @param GhAddonOrderItem $orderItem Bestellposition
   * @param GhAddonOrder $order Auftrag
   * @return string
   */
  public function getProductClickHandler($orderItem, $order)
  {
    return null;
  }
}