<?php
/*
 * GREYHOUND Addon Framework
 *
 * Copyright (c) 2014-2019 GREYHOUND Software GmbH & Co. KG
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of the GREYHOUND Addon Framework and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including without
 * limitation the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to permit persons to
 * whom the Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/**
 * AJAX Proxy Klasse zum abrufen von Daten aus Drittsystemen.
 *
 * @package GhAddonFramework
 * @subpackage page
 * @category PHP Addon Framework
 *
 * @author digital guru GmbH &amp; Co. KG <develop@greyhound-software.com>
 * @copyright 2014-2019 digital guru GmbH &amp; Co. KG
 * @link greyhound-software.com
 */
class GhAddonAjax
{
  /**
   * Addon.
   * @var GhAddon
   */
  protected $_addon;

  /**
   * Im Falle einer JSONRPC-Anfrage die korrespondierende Id, ansonsten 0.
   */
  protected $_rpcId = 0;

  /**
   * Cache-Datei-Handle.
   * @var resource
   */
  protected $_cache;

  /**
   * Zeit, nach der Cache Einträge ungültig werden (in Sekunden).
   * @var int
   */
  protected $_cacheMaxAge = 10;

  /**
   * Flag, ob der Cache verwendet werden soll.
   * @var boolean
   */
  protected $_useCache;

  /**
   * Cache-Schlüssel.
   * @var string
   */
  protected $_cacheKey;


  /**
   * @param GhAddon $addon Addon
   */
  public function __construct($addon)
  {
    $this->_addon = $addon;
  }

  /**
   * Sicherstellen, dass das Addon mit gültigen GREYHOUND Benutzerdaten ausgeführt wird.
   */
  protected function _authenticate()
  {
    // Die UserInfo wird von der getGreyhoundUserInfo Funktion gecached, so dass hier nicht ständig Netzwerkzugriffe stattfinden:
    try {
      $userInfo = $this->_addon->getGreyhoundUserInfo(false, true);

      if (!$userInfo) {
        throw new Exception('Es konnten keine Benutzerinformationen ermittelt werden.');
      }
    } catch(Exception $e) {
      $host = $this->_addon->getGreyhoundClient()->getHost();
      $host = $host ? ' (Verwendeter Host/IP: ' . $host . ')' : '';
      throw new GhAddonConnectionException('Es konnte keine Verbindung zum GREYHOUND Server hergestellt werden. Bitte überprüfen Sie die Einstellung "Öffentlicher Host/IP" des GREYHOUND Servers.' . $host);
    }
  }

  /**
   * Verarbeitet den Request und liefert das Ergebnis zurück.
   */
  protected function _processRequest()
  {
    GhCoreGlobals::continueSession(get_class($this->_addon));

    if(!GhCoreGlobals::session('settingsTimestamp'))
      GhCoreGlobals::session('settingsTimestamp', time(), true);

    $requestSettingsTimestamp = GhCoreGlobals::getValue('ghsettingstimestamp', 0);

    if($requestSettingsTimestamp && $requestSettingsTimestamp != GhCoreGlobals::session('settingsTimestamp'))
      throw new GhAddonSettingsChangedException('settings timestamp changed', 409, GhCoreGlobals::session('settingsTimestamp'));

    return $this->_processRpcRequest();
  }

  /**
   * Liefert Daten, die für die Vorschauzeile abgerufen wurden.
   *
   * @throws GhAddonException falls keine Felder für die Vorschauzeile konfiguriert wurden.
   *
   * @param array $params RPC Parameter
   * @return array
   */
  protected function _processPreviewLineRequest($params)
  {
    $previewLineFields = $this->_addon->getPreviewLineFields();

    if(empty($previewLineFields))
      throw new GhAddonException('In den Einstellungen des Addons wurden noch keine Felder für die Vorschauzeile konfiguriert.');

    $accountId = $params[0];
    $orderId   = $params[1];
    $orderData = isset($params[2]) ? $params[2] : null;
    $hash = isset($params[3]) ? $params[3] : null;

    $account = $this->_addon->getAccount($accountId);

    if(!$account)
      throw new GhAddonConnectionException('Kein passender Account konfiguriert');

    $data = $account->newOrderData($orderData);
    $order = null;
    $customer = null;

    if($data)
    {
      $order = $data->getOrder($orderId);

      if($order && $order->CustomerID)
        $customer = $data->getCustomer($order->CustomerID);
    }

    if(!$order)
      return array();

    $result = array();
    $variables = $account->getTemplateVariables($customer, $order, null);

    // Felder für die Vorschauzeile:
    foreach($previewLineFields as $previewLineField)
    {
      if(isset($variables[$previewLineField['Var']]))
      {
        $value = $variables[$previewLineField['Var']];
        $value = $this->_processTemplateVariable($previewLineField['Var'], $value, $order, $customer, $account, $variables);

        if(strlen($previewLineField['Caption']) > 0 || strlen($value) > 0)
          $result[] = (object)array('key' => $previewLineField['Caption'], 'value' => $value);
      }
    }

    // Vorgangsnummer setzen:
    $taskNumberPrefix = $this->_addon->getAddonSetting('TaskNumberPrefix', null);
    $assignUserFieldsTaskNumberPrefixes = $this->_addon->getAddonSetting('AssignUserFieldsTaskNumberPrefixes', null);
    $taskNumberAutoAssign = $this->_addon->getAddonSetting('TaskNumberAutoAssign', null);
    $orderTaskNumber = $account->orderIdToTaskNumber($orderId);

    if($orderId && $taskNumberPrefix != "" && $taskNumberPrefix != "0" && $hash['itemID'] && $hash['itemTaskNumber'] && $hash['itemTaskNumber'] != $orderTaskNumber)
    {
      // Die Vorgangsnummer wird automatisch gesetzt, sofern dies in den Einstellungen so festgelegt wurde:
      if($hash['itemKind'] != rpc_itemslib::ikContact && $taskNumberAutoAssign === "1" && $hash['itemTaskNumber'] != $orderTaskNumber && ($hash['itemTaskNumber'][0] == 'T' || ($hash['itemTaskNumber'][0] == $taskNumberPrefix)))
      {
        $this->_setTaskNumber(array($hash['itemID'], $orderTaskNumber));
        $hash['itemTaskNumber'] = $orderTaskNumber;
      }
    }

    $prefixes = $assignUserFieldsTaskNumberPrefixes ? explode(',', $assignUserFieldsTaskNumberPrefixes) : array();

    for ($i = 0; $i < count($prefixes); $i++) {
      $prefixes[$i] = strtoupper(trim($prefixes[$i]));
    }

    // Datenübernahme:
    if($orderId && $hash['itemID'] && (
      (($taskNumberPrefix == "0" || $taskNumberPrefix == 'T') && $hash['itemTaskNumber'] && $hash['itemTaskNumber'][0] == 'T') ||
      (in_array('*', $prefixes) || ($hash['itemTaskNumber'] && in_array($hash['itemTaskNumber'][0], $prefixes))) ||
      ($hash['itemTaskNumber'] && $hash['itemTaskNumber'][0] == $taskNumberPrefix)
      )) {
      $this->_setUserFields(array($accountId, $hash['itemID'], $orderId, $orderData));
    }

    return $result;
  }

  /**
   * Liefert Daten, die für die Seitenleiste abgerufen wurden.
   * Das Ergebnis ist ein Objekt mit den Eigenschaften "customers", "orders" und "selectedOrder".
   *
   * @throws Exception falls ein Fehler beim Abrufen der Daten auftritt
   * @throws GhAddonConnectionException falls ein Verbindungsfehler auftritt oder kein passender Account konfiguriert ist
   *
   * @return mixed
   */
  protected function _processRpcRequest()
  {
    $rpcRequest = json_decode(file_get_contents('php://input'), true);
    $method = $rpcRequest['method'];
    $params = $rpcRequest['params'];
    $this->_useCache = !isset($rpcRequest['cache']) || !$rpcRequest['cache'] ? true : false;
    $uncache = isset($rpcRequest['uncache']) && $rpcRequest['uncache'];
    $this->_cacheKey = null;
    $this->_cache = null;
    $this->_rpcId = $rpcRequest['id'];

    if (!$this->_rpcId) {
      $this->_rpcId = 0;
    }

    if ($this->_useCache) {
      $this->_cacheKey = get_class($this->_addon) . '-' . md5($method . '#' . json_encode($params));
      $this->_clearCache($uncache ? 0 : $this->_cacheMaxAge);
      $response = $this->_readCache($this->_cacheKey, $this->_cacheMaxAge);

      if ($response) {
        $result = json_decode($response, JSON_OBJECT_AS_ARRAY);

        if (is_array($result) && isset($result['addon']) && $result['addon'] == get_class($this->_addon)) {
          if ($result['success']) {
            return $result['data'];
          } else {
            $exceptionClass = $result['data']['class'];

            if (!class_exists($exceptionClass)) {
              $exceptionClass = 'GhAddonException';
            }

            $e = new $exceptionClass($result['data']['message'], $result['data']['code']);

            if ($exceptionClass == 'GhAddonEmptySearchException') {
              /* @var GhAddonEmptySearchException $e */
              $e->setSearchTerm($result['data']['searchTerm']);
            }

            throw $e;
          }
        }
      }
    }

    $result = null;
    $error = null;

    try {
      $result = $this->_processRpcMethod($method, $params);
    } catch(Exception $exception) {
      $error = $exception;
    }

    if ($this->_cache) {
      if ($error) {
        $data = array('addon' => get_class($this->_addon), 'success' => false, 'data' => array('class' => get_class($error), 'message' => $error->getMessage(), 'code' => $error->getCode()));

        if (get_class($error) == 'GhAddonEmptySearchException') {
          /* @var GhAddonEmptySearchException $error */
          $data['data']['searchTerm'] = $error->getSearchTerm();
        }
      } else {
        $data = array('addon' => get_class($this->_addon), 'success' => true, 'data' => $result);
      }

      @$this->_endCacheWrite($this->_cache, json_encode($data));
    }

    if ($error) {
      if ($this->_addon && !isset($error->dontLog)) {
        if (get_class($error) == 'GhAddonEmptySearchException') {
          $this->_addon->log('AJAX: ' . $error->getMessage());
        } else {
          $this->_addon->log('AJAX Fehler: ' . $error->getMessage());
        }
      }
      
      throw $error;
    } else {
      return $result;
    }
  }

  /**
   * Verarbeitet eine RPC Methode.
   *
   * @throws GhAddonCacheLockedException falls die Cache-Datei gesperrt ist und nicht geschrieben werden kann
   * @throws GhAddonConnectionException falls ein Fehler bei der Verbindung zum Drittanbieter aufgetreten ist
   * @throws GhAddonException falls ein sonstiger Fehler aufgetreten ist
   *
   * @param string $method Methode
   * @param mixed $params Parameter
   * @return mixed
   */
  protected function _processRpcMethod($method, $params)
  {
    switch($method)
    {
      case 'previewline':
        if($this->_useCache && $this->_cacheKey)
          $this->_beginCacheWrite($this->_cacheKey);

        return $this->_processPreviewLineRequest($params);

      case 'search':
        $debugFile = $this->_addon->getDebugLogFile();

        if($debugFile && file_exists($debugFile))
          unlink($debugFile);

        if($this->_useCache && $this->_cacheKey)
          $this->_beginCacheWrite($this->_cacheKey);

        return $this->_search($params);

      case 'getAddonInfo':
        return $this->_getAddonInfo($params);

      case 'getAddonHelp':
        return $this->_getAddonHelp($params);

      case 'getAccounts':
        return $this->_getAccounts($params);

      case 'getTemplates':
        if($this->_useCache && $this->_cacheKey)
          $this->_beginCacheWrite($this->_cacheKey);

        return $this->_getTemplates($params);

      case 'getItem':
        if($this->_useCache && $this->_cacheKey)
          $this->_beginCacheWrite($this->_cacheKey);

        return $this->_getItem($params);

      case 'getOrderDetails':
        if($this->_useCache && $this->_cacheKey)
          $this->_beginCacheWrite($this->_cacheKey);

        return $this->_getOrderDetails($params);

      case 'getAddonSettings':
        return $this->_getAddonSettings($params);

      case 'getAddonUserSettings':
        return $this->_getAddonUserSettings($params);

      case 'setAddonUserSettings':
        return $this->_setAddonUserSettings($params);

      case 'setAddonUserSetting':
        return $this->_setAddonUserSetting($params);

      case 'getNewAccountSettings':
        return $this->_getNewAccountSettings();

      case 'setTaskNumber':
        return $this->_setTaskNumber($params);

      case 'setUserFields':
        return $this->_setUserFields($params);

      //  Settings-Calls

      case 'getAddonSettingsOptions':
        return $this->_getAddonSettingsOptions($params);

      case 'getAddonTemplateVariables':
        return $this->_getAddonTemplateVariables($params);

      case 'createUserIdUserField':
        return $this->_createUserIdUserField($params);

      case 'checkConnection':
        return $this->_checkConnection($params);

      case 'saveSettings':
        return $this->_saveSettings($params);

      case 'getContactRef':
        return $this->_getContactRef($params);

      case 'sendDebugLog':
        return $this->_sendDebugLog($params);
      
      case 'log':
        $this->_addon->log(json_encode($params));
        return null;

      default:
        throw new GhAddonException('Ungültige RPC Methode: ' . $method);
    }
  }

  /**
   * Sucht Auftragsdaten anhand eines Suchbegriffes und liefert die gefundenen Aufträge und Kunden zurück.
   *
   * @throws GhAddonConnectionException falls keine Verbindung konfiguriert ist oder die Verbindung nicht hergestellt werden kann
   * @throws GhAddonException falls ein Fehler beim Datenabruf auftritt
   *
   * @param array $params RPC Parameter
   * @return object
   */
  protected function _search($params)
  {
    $accountId = $params[0];
    $account = $this->_addon->getAccount($accountId);

    if (!$account) {
      throw new GhAddonConnectionException('Kein passender Account konfiguriert');
    }

    $searchType = $params[1];
    $searchTerm = $params[2];
    $itemParams = $params[3];

    if ($searchType == 'searchBySearchTerm' && !$searchTerm) {
      throw new GhAddonException('Kein Suchbegriff übergeben!');
    }

    $result = null;
    $data = null;
    $searchedTerm = '';
    $searchedFields = array();
    $searchedInfo = '';

    if ($searchTerm) {
      try {
        switch ($searchType) {
          case 'search':
            $searchedTerm = $searchTerm;

            if (is_array($searchTerm) && count($searchTerm) == 1) {
              foreach ($searchTerm as $key => $value) {
                if ($key == 'taskNumber' && $value) {
                  $searchedTerm = $account->taskNumberToOrderId($value);
                }
              }
            }

            $data = $account->search(null, null, $searchTerm);

            if (!$data && GhAddonData::safeGet($itemParams, 'itemKind') == rpc_itemslib::ikContact) {
              $itemId = GhAddonData::safeGet($itemParams, 'itemId');
              $userFieldId = $this->_addon->getAddonSetting('CustomerIDUserField', 0);

              if ($userFieldId) {
                $itemService = rpc_itemslib::newRpcItems($this->_addon->getGreyhoundClient());
                $filter = rpc_itemslib::newRpcItemFilter();
                $filter->ItemRefs = array($itemId);
                $items = $itemService->GetList($filter, rpc_itemslib::iobCreated, rpc_itemslib::iodAscending, 0, 0, array($userFieldId), 0);
                $customerId = GhAddonData::safeGet($items, '0/UserValues/0/Value');

                if ($customerId) {
                  $searchedTerm = '' . $customerId;
                  $data = $account->searchByCustomerIds(array($customerId), $params);
                }
              }
            }

            break;
          case 'searchByEmail':
            $searchedTerm = $searchTerm;
            $data = $account->searchByEmail($searchTerm, $params);
            break;
          case 'searchBySearchTerm':
            $searchedTerm = $searchTerm;
            $data = $account->searchBySearchTerm($searchTerm, $params);
            break;
          case 'searchByEbayName':
            $searchedTerm = $searchTerm;
            $data = $account->searchByEbayName($searchTerm, $params);
            break;
          case 'searchByCustomerIds':
            $searchedTerm = $searchTerm;
            $data = $account->searchByCustomerIds($searchTerm, $params);
            break;
          case 'searchByOrderIds':
            $searchedTerm = $searchTerm;
            $data = $account->searchByOrderIds($searchTerm, $params);
            break;
          case 'searchByTaskNumber':
            $searchedTerm = $account->taskNumberToOrderId($searchTerm);
            $data = $account->searchByTaskNumber($searchTerm, $params);
            break;
          default:
            $data = null;
        }
      }
      catch (GhAddonEmptySearchException $e) {
        $searchedFields = array_merge($searchedFields, $e->getSearchedFields());
        $searchedInfo = $e->getSearchInfo();

        if (strlen($searchedTerm) == 0) {
          $searchedTerm = $e->getMessage();
        }
      }
    }

    // Nach Kunden-ID oder Kundennummer eines klassifizierten Kontakts suchen (falls vorhanden):
    if (!$data && $searchType != 'searchBySearchTerm' && isset($params[3]) && is_array($params[3])) {
      try {
        $itemId = isset($params[3]['itemId']) ? $params[3]['itemId'] : null;
        $itemKind = isset($params[3]['itemKind']) ? $params[3]['itemKind'] : null;
        $itemClassified = isset($params[3]['itemClassified']) ? $params[3]['itemClassified'] : array();

        if (!is_array($itemClassified)) {
          $itemClassified = array($itemClassified);
        }

        if ($itemId && $itemKind !== null) {
          $customerId = $this->_addon->getCustomerIdFromUserField($itemId, $itemKind, $itemClassified);

          if ($customerId) {
            $searchedTerm = '' . $customerId;
            $data = $account->searchByCustomerIds(array($customerId));
          } else {
            $customerNumber = $this->_addon->getCustomerNumberFromUserField($itemId, $itemKind, $itemClassified);

            if ($customerNumber) {
              $searchedTerm = '' . $customerNumber;
              $data = $account->searchByCustomerNumbers(array($customerNumber));
            } else {
              $emails = $this->_addon->getEmailsFromContacts($itemClassified);

              if ($emails) {
                foreach ($emails as $email) {
                  $searchedTerm = $email;
                  $data = $account->searchByEmail($email);

                  if ($data) {
                    break;
                  }
                }
              }
            }
          }
        }
      } catch (GhAddonEmptySearchException $e) {
        $searchedTerm = $e->getSearchTerm();
        $searchedFields = array_merge($searchedFields, $e->getSearchedFields());
        $searchedInfo = $e->getSearchInfo();

        if (strlen($searchedTerm) == 0) {
          $searchedTerm = $e->getMessage();
        }
      }
    }

    // Prüfen, ob am Element Auftrags- oder Kundennummern in benutzerdefinierten Feldern gespeichert sind, und dann ggf.
    // danach suchen:
    if (!$data && $searchType != 'searchBySearchTerm') {
      try {
        $itemId = isset($params[3]['itemId']) ? $params[3]['itemId'] : null;

        if ($itemId) {
          $data = $this->_searchByUserFields($account, $itemId);

          if ($data && $data->getSearchTerm()) {
            $searchedTerm = $data->getSearchTerm();
          }
        }
      } catch (GhAddonEmptySearchException $e) {
        $searchedTerm = $e->getSearchTerm();
        $searchedFields = array_merge($searchedFields, $e->getSearchedFields());
        $searchedInfo = $e->getSearchInfo();

        if (strlen($searchedTerm) == 0) {
          $searchedTerm = $e->getMessage();
        }
      }
    }

    if (is_object($searchedTerm)) {
      /* @var object $searchedTerm */
      $searchedTerm = get_object_vars($searchedTerm);
    }

    if (is_array($searchedTerm)) {
      foreach ($searchedTerm as $key => $value) {
        if ($value) {
          $searchedTerm = $value;
          break;
        }
      }
    }

    if ($data) {
      $debugFile = $this->_addon->getDebugLogFile();

      if (!file_exists($debugFile)) {
        $debugFile = null;
      }

      $result = (object)array('customers' => array(), 'orders' => array(), 'selectedOrderId' => null, 'data' => $data->getData(), 'searchedTerm' => $searchedTerm, 'debugFile' => $debugFile);

      foreach ($data->getCustomers() as $customer) {
        $result->customers[] = $customer;
      }

      foreach ($data->getOrders() as $order) {
        $result->orders[] = $order;
      }

      if ($data->getSelectedOrderId()) {
        $result->selectedOrderId = $data->getSelectedOrderId();
      } elseif (count($result->orders) > 0) {
        $result->selectedOrderId = $result->orders[0]->ID;
      }

      //  Details für selektierte Bestellung abrufen; ansonsten orderTaskNumber patchen
      foreach ($result->orders as $key=>$order) {
        if ($order->ID == $result->selectedOrderId) {
          $result->orders[$key] = $data->getOrder($result->selectedOrderId);
        }

        // Patchen der Vorgangsnummer:
        if ($result->orders[$key]->TaskNumberID) {
          $result->orders[$key]->OrderTaskNumberID = $account->orderIdToTaskNumber($result->orders[$key]->TaskNumberID);
        } else {
          $result->orders[$key]->OrderTaskNumberID = $account->orderIdToTaskNumber($result->orders[$key]->ID);
        }

        // ShortName setzen:
        $result->orders[$key]->ShortName = trim(($order->Date ? $order->Date->getFormatted('d.m.Y') : '') . ', ' . (GhAddonData::safeGet($order->BillingAddress, 'Company') ? GhAddonData::safeGet($order->BillingAddress, 'Company') : GhAddonData::safeGet($order->BillingAddress, 'Name')), ' ,');
      }

      // Beim Abrufen der Auftragsdetails wurden evtl. die Daten noch angereichert:
      $result->data = $data->getData();
    }
    
    if ($data) {
      return $result;
    } else {
      if (strlen($searchedTerm) > 0) {
        $message = 'Die Suche nach "' . $searchedTerm . '"';
      } else {
        $message = 'Die Suche';
      }

      $searchedFields = array_unique($searchedFields);

      if (!empty($searchedFields)) {
        if (count($searchedFields) == 1) {
          $message .= ' im Feld ' . $searchedFields[0];
        } else {
          $message .= ' in den Feldern ' . $searchedFields[0];

          for ($i = 1; $i < count($searchedFields) - 1; $i++) {
            $message .= ', ' . $searchedFields[$i];
          }

          $message .= ' und ' . $searchedFields[count($searchedFields) - 1];
        }
      }

      $message .= ' lieferte keine Ergebnisse.';

      if ($searchedInfo) {
        $message .= ' ' . $searchedInfo;
      }

      throw new GhAddonEmptySearchException($message, $searchedFields, $searchedTerm);
    }
  }

  /**
   * Sucht anhand von benutzerdefinierten Feldern im Drittsystem nach Aufträgen.
   *
   * @param GhAddonAccount $account Account
   * @param int $itemId Element-ID
   * @return GhAddonData
   */
  protected function _searchByUserFields($account, $itemId)
  {
    $userFieldIds = array();
    $searchTemplateVariables = $this->_addon->getSearchTemplateVariables();

    foreach ($this->_addon->getUserFieldVars() as $var) {
      $value = trim($var['field']);

      if (in_array($var['var'], $searchTemplateVariables) && $value && !in_array($value, $userFieldIds)) {
        $userFieldIds[] = $value;
      }
    }

    $customerNoUserField = $this->_addon->getAddonSetting('CustomerNoUserField');

    if ($customerNoUserField && !in_array($customerNoUserField, $userFieldIds)) {
      $userFieldIds[] = $customerNoUserField;
    }

    if (!empty($userFieldIds)) {
      $userFieldSearchTerms = array();
      $filter = rpc_itemslib::newRpcItemFilter();
      $filter->ItemRefs = array($itemId);

      foreach (rpc_itemslib::newRpcItems($this->_addon->getGreyhoundClient())->GetList($filter, rpc_itemslib::iobCreated, rpc_itemslib::iodAscending, 0, 0, $userFieldIds, 0) as $item) {
        if (isset($item->UserValues) && is_array($item->UserValues)) {
          /** @var GhRpcItemslibRpcListItem $item */
          foreach ($item->UserValues as $userValue) {
            $value = trim($userValue->Value);

            if (in_array($userValue->FieldRef, $userFieldIds) && !in_array($value, $userFieldSearchTerms)) {
              $userFieldSearchTerms[] = $value;
            }
          }
        }
      }

      foreach ($userFieldSearchTerms as $userFieldSearchTerm) {
        $data = $account->search(null, null, array('search' => $userFieldSearchTerm));

        if ($data) {
          $data->setSearchTerm($userFieldSearchTerm);

          return $data;
        }
      }
    }

    return null;
  }

  /**
   * Liefert die Accounts die für das Addon konfiguriert sind.
   *
   * @param array $params RPC Parameter
   * @return array
   */
  protected function _getAccounts($params)
  {
    $accounts = array();

    foreach($this->_addon->getAccounts() as $account)
      $accounts[] = (object)array(
        'id' => $account->getId(),
        'isValid' => $account->isValid() ? true : false,
        'isActive' => $account->isActive() ? true : false,
        'recipientFilters' => $account->getRecipientFilters(),
        'settings' => $account->getSettings()
      );

    return $accounts;
  }

  /**
   * Liefert eine Liste der Templates (GREYHOUND Textbausteine).
   *
   * @param array $params RPC Parameter
   * @return array
   */
  protected function _getTemplates($params)
  {
    $templates = array();

    $accountId = $params[0];
    $account = $this->_addon->getAccount($accountId);

    foreach($account->getTemplates(true) as $template)
    {
      $templates[] = (object)array(
        'id' => $template->ID,
        'ParentRef' => $template->ParentRef,
        'Name' => $template->Name,
        'Subject' => $template->Subject,
        'isHTML' => $template->IsHtml,
        'Description' => $template->Description
      );
    }

    return $templates;
  }

  /**
   * Liefert Item aus GH
   *
   * @param array $params RPC Parameter
   * @return array
   */
  protected function _getItem($params)
  {
    $item = array('isHTML' => false, 'content' => null, 'subject' => null, 'signature' => null, 'templateVars' => array());

    $accountId= $params[0];
    $account  = $this->_addon->getAccount($accountId);

    $itemId   = $params[1];
    $kind     = $params[2];
    $isReply  = $params[3];
    $itemGroup= $params[4];

    $orderId   = $params[5];
    $orderData = isset($params[6]) ? $params[6] : null;
    $templateId = $params[7];

    $data = $account->newOrderData($orderData);
    $order = null;
    $customer = null;

    if($data)
    {
      $order = $data->getOrder($orderId);

      if($order && $order->CustomerID)
        $customer = $data->getCustomer($order->CustomerID);
    }

    $template = $account->getTemplate($templateId);

    if($itemId) // Antwort / Weiterleiten:
    {
      $newItem = rpc_itemslib::newRpcItems($this->_addon->getGreyhoundClient())->GetAs($itemId, $kind, $isReply ? rpc_itemslib::insAnswer : rpc_itemslib::insForward, false);

      $format = $this->_addon->getItemFormat($kind, $isReply ? rpc_itemslib::insAnswer : rpc_itemslib::insForward, GhRpcUtils::inSet($newItem->Flags, rpc_itemslib::ifUseHtml));

      $item['subject']  = $newItem->Subject;

      // WM_COPYDATA Inhalte werden in UTF-8 erwartet:
      $item['content']  = iconv('windows-1252', 'UTF-8', $newItem->Properties->Email->EditorData);
    }
    else // Neues Element:
    {
      $format = $this->_addon->getItemFormat($kind, rpc_itemslib::insNew, $template ? $template->IsHtml : null);
      $item['subject'] = '';
      $item['content'] = '';
    }

    $item['isHTML'] = ($format == 'html');

    if($itemGroup)
    {
      $item['signature'] = $this->_addon->getSignature($itemGroup, $format);

      // Die Plain-Text Signatur wird vom Server (fälschlicherweise) bereits in UTF-8 geliefert, die HTML Signatur muss noch konvertiert werden:
      if($item['isHTML'])
        $item['signature'] = iconv('windows-1252', 'UTF-8', $item['signature']);
    }

    if($item['signature'] && !$item['isHTML'])
      $item['signature'] = preg_replace('/([^\r])\n/', '$1' . "\r\n", $item['signature']);  // Defekte Zeilenumbrüche korrigieren:

    if($item['signature'] && $item['isHTML'])
      $item['signature'] = '<p>&nbsp;</p>' . "\r\n" . '<div class="signature">' . GhCoreStrings::getHtmlBody($item['signature']) . '</div>' . "\r\n";

    $item['templateVars'] = $account->getTemplateVariables($customer, $order, $template);
    $item['template'] = $account->applyTemplate($template, $order, $customer, $format);

    return $item;
  }

  /**
   * Liefert Detaildaten einer Bestellung zurück, oder null, falls es keine zusätzlichen Daten zur Bestellung gibt.
   *
   * @throws GhAddonConnectionException falls kein passender Account konfiguriert wurde
   *
   * @param array $params RPC Parameter
   * @return object
   */
  protected function _getOrderDetails($params)
  {
    $accountId = $params[0];
    $orderId   = $params[1];
    $orderData = isset($params[2]) ? $params[2] : null;

    $account = $this->_addon->getAccount($accountId);

    if(!$account)
      throw new GhAddonConnectionException('Kein passender Account konfiguriert');

    // Da im JavaScript die Kundendaten sowie Stammdaten der Bestellung bereits vorhanden sind,
    // reicht es hier aus mit einer "leeren" Bestellung nur die zusätzlichen Details abzurufen (getOrder).
    $data = $account->newOrderData($orderData);

    if(!$data)
      return null;

    $order = $data->getOrder($orderId);

    // Setzen der Vorgangsnummer:
    if(!empty($order) && $order->ID)
    {
      if($order->TaskNumberID)
        $order->OrderTaskNumberID = $account->orderIdToTaskNumber($order->TaskNumberID);
      else
        $order->OrderTaskNumberID = $account->orderIdToTaskNumber($order->ID);
      

      $debugFile = $this->_addon->getDebugLogFile();

      if (file_exists($this->_addon->getDebugLogFile()))
        $order->DebugFile = $debugFile;
      else
        $order->DebugFile = null;
    }

    return $order;
  }

  /**
   * Liefert die Einstellungen zu diesem Addon zurück
   *
   * @param array $params RPC Parameter
   * @return array
   */
  protected function _getAddonSettings($params)
  {
    $result = array();

    foreach ($this->_addon->getAddonSettings(true) as $key => $value)
      $result[] = (object)array('key' => $key, 'value' => $value);

    return $result;
  }

  /**
   * Liefert die Benutzer-Einstellungen des aktuellen Benutzers für dieses Addon zurück.
   *
   * @param array $params RPC Parameter
   * @return array
   */
  protected function _getAddonUserSettings($params)
  {
    return $this->_addon->getAddonUserSettings();
  }

  /**
   * Setzt die Benutzer-Einstellungen für dieses Addon für den aktuellen Benutzer.
   *
   * @param array $params RPC Parameter
   * @return array
   */
  protected function _setAddonUserSettings($params)
  {
    return $this->_addon->setAddonUserSettings($params[0]);
  }

  /**
   * Setzt eine Benutzer-Einstellung für dieses Addon für den aktuellen Benutzer.
   *
   * @param array $params RPC Parameter
   * @return array
   */
  protected function _setAddonUserSetting($params)
  {
    return $this->_addon->setAddonUserSetting($params[0], $params[1]);
  }

  /**
   * Liefert die Einstellungen für einen neuen Account.
   *
   * @return array
   */
  protected function _getNewAccountSettings()
  {
    return $this->_addon->getNewAccountSettings();
  }

  /**
   * Liefert die Addon Einstellungen und die Optionen für den Einstellungsdialog.
   *
   * @param array $params RPC Parameter
   * @return array
   */
  protected function _getAddonSettingsOptions($params)
  {
    $templateVars = $this->_addon->getTemplateVariableDescriptions();

    $settings = array();

    foreach ($this->_addon->getAddonSettings(true) as $key => $value)
      $settings[] = (object)array('key' => $key, 'value' => $value);


    return array(
      'pages' => $this->_addon->getOptionPages(),
      'settings' => $settings,
      'account' => $this->_fixAddonOptions($this->_addon->getAccountOptions()),
      'global' => $this->_fixAddonOptions($this->_addon->getGlobalOptions()),
      'extended' => $this->_fixAddonOptions($this->_addon->getExtendedOptions()),
      'accountDefaults' => $this->_addon->getNewAccountSettings(),
      'templateVariableDescriptions' => $templateVars,
      'userDefinedFields' => rpc_userfieldslib::newRpcUserFields($this->_addon->getGreyhoundClient())->GetList()
    );
  }

  /**
   * Liefert die Template-Variablen des Addons.
   *
   * @param array $params RPC Parameter
   * @return array
   */
  protected function _getAddonTemplateVariables($params)
  {
    $vars = array();

    foreach($this->_addon->getTemplateVariableDescriptions() as $var => $description)
      $vars[] = array('name' => $var, 'desc' => $description);

    return $vars;
  }

  /**
   * Liefert Basisinformationen über das Addon.
   *
   * @param array $params RPC Parameter
   * @return array
   */
  protected function _getAddonInfo($params)
  {
    return array(
      'name' => $this->_addon->getName(),
      'version' => $this->_addon->getVersion(),
      'build' => $this->_addon->getBuild(),
      'remoteServiceName' => $this->_addon->getRemoteServiceName(),
      'searchTooltip' => $this->_addon->getSearchTooltip(),
      'emptySearchMessage' => 'Die Suche lieferte keine Ergebnisse.',
      'sidebarExtensionIdentifier' => $this->_addon->getSidebarIdentifier(),
      'settingsExtensionIdentifier' => $this->_addon->getSettingsIdentifier(),
      'webExtensionIdentifier' => $this->_addon->getWebIdentifier(),
      'helpExtensionIdentifier' => $this->_addon->getHelpIdentifier(),
      'settingsTimestamp' => GhCoreGlobals::session('settingsTimestamp')
    );
  }

  /**
   * Liefert den individuellen Hilfetext des Addons für den Hilfe-Dialog.
   * 
   * @param array $params RPC Parameter
   * @return string
   */
  protected function _getAddonHelp($params)
  {
    return $this->_addon->getHelp();
  }

  /**
   * Erzeugt ein benutzerdefiniertes Feld für die Kunden-ID des Drittanbieters und liefert die ID des neuen
   * benutzerdefinierten Felds.
   *
   * @param array $params RPC Parameter
   * @return int
   */
  protected function _createUserIdUserField($params)
  {
    return $this->_addon->createCustomerIdUserField();
  }

  /**
   * Prüft Verbindungsdaten
   *
   * @param array $params RPC Parameter (Verbindungsdaten)
   * @return array
   */
  protected function _checkConnection($params)
  {
    if(empty($params))
      return 'Invalid params';

    return $this->_addon->checkConnection($params[0]);
  }

  /**
   * Speichert die Einstellungen und liefert den neuen Settings-Timestamp zurück.
   *
   * @param array $settings Einstellungen
   * @return int
   */
  protected function _saveSettings($settings)
  {
    $this->_addon->setAddonSettings($settings);

    foreach($this->_addon->getAccounts() as $account)
      $account->clearCache();

    $this->_clearCache();

    GhCoreGlobals::session('settingsTimestamp', time(), true);

    return GhCoreGlobals::session('settingsTimestamp');
  }

  /**
   * Korrigiert Optionen für den Einstellungsdialog, so dass diese besser JSON-codiert werden können.
   *
   * @param array $options Optionen für den Einstellungsdialog
   * @return array
   */
  protected function _fixAddonOptions($options)
  {
    foreach($options as &$option)
    {
      if(isset($option['options']) && ($option['options'] instanceof GhCoreTree))
      {
        $option['options'] = $this->_fixTree($option['options']);
      }
    }

    return $options;
  }

  /**
   * Wandelt einen GhCoreTree in eine simple Objektstruktur um, die sich besser JSON-codieren lässt.
   *
   * @param GhCoreTreeBase $tree Baumstruktur oder Baumknoten
   * @return array|null
   */
  protected function _fixTree($tree)
  {
    $children = array();

    foreach($tree->getChildren() as $child)
      $children[] = $this->_fixTree($child);

    if($tree instanceof GhCoreTree)
      $node = array('id' => 0, 'caption' => '', 'data' => null, 'children' => $children);
    elseif($tree instanceof GhCoreTreeNode)
      $node = array('id' => $tree->getId(), 'caption' => $tree->getCaption(), 'data' => $tree->getData(), 'children' => $children);
    else
      $node = null;

    return $node;
  }

  /**
   * Setzt die Vorgangsnummer.
   *
   * @throws GhAddonException falls die Vorgangsnummer nicht gesetzt werden konnte
   *
   * @param array $params RPC Parameter
   * @return array
   */
  protected function _setTaskNumber($params)
  {
    $itemId = $params[0];
    $newTaskNumber = $params[1];

    if($itemId && $newTaskNumber)
    {
      try
      {
        $this->_addon->setItemTaskNumber($itemId, $newTaskNumber);
      }
      catch(Exception $exception)
      {
        throw new GhAddonException('Vorgangsnummer konnte nicht gesetzt werden. (' . $exception->getMessage() . ')', $exception->getCode());
      }
    }
    return null;
  }

  /**
   * Übernimmt Auftragsdaten in benutzerdefinierte Felder eines Elements.
   *
   * @throws GhAddonConnectionException falls kein passender Account vorliegt
   * @throws GhAddonException falls ein Fehler bei der Datenübernahme aufgetreten ist
   *
   * @param array $params RPC Parameter
   * @return boolean
   */
  protected function _setUserFields($params)
  {
    $accountId = $params[0];
    $itemId = $params[1];
    $orderId   = $params[2];
    $orderData = isset($params[3]) ? $params[3] : null;
    $error = null;

    $account = $this->_addon->getAccount($accountId);

    if(!$account)
      throw new GhAddonConnectionException('Kein passender Account konfiguriert');

    $data = $account->newOrderData($orderData);

    if(!$data)
      throw new GhAddonException('Keine Daten für die Datenübernahme vorhanden');

    $order = $data->getOrder($orderId);

    if(!$order)
      throw new GhAddonException('Ungültige Bestellung (Order-ID nicht gefunden)');

    $customer = $data->getCustomer($order->CustomerID);

    $userFieldVars = $this->_addon->getUserFieldVars();

    if(!empty($userFieldVars) && $itemId)
    {
      $vars = $account->getTemplateVariables($customer, $order, null);
      $itemService = rpc_itemslib::newRpcItems($this->_addon->getGreyhoundClient());

      $itemChanged = false;
      $item = $itemService->Get($itemId, false);
      $item->Protocol = array();
      $item->Remarks = array();

      if(!is_array($item->UserFields))
        $item->UserFields = array();

      // Benutzerdefinierte Felder befüllen:
      foreach($userFieldVars as $userFieldVar)
      {
        $userFieldId = $userFieldVar['field'];

        foreach($item->UserFields as &$tmpUserField)/* @var GhRpcItemslibRpcUserFieldItem $tmpUserField */
        {
          if($tmpUserField->FieldRef == $userFieldId)
          {
            $tmpValue = isset($vars[$userFieldVar['var']]) ? $vars[$userFieldVar['var']] : '';
            $userFieldValue = $this->_getUserFieldValue($tmpValue, $tmpUserField->DataType, $userFieldVar['var']);

            if($userFieldValue != $tmpUserField->Value)
              $itemChanged = true;

            $tmpUserField->Value = $userFieldValue;
            break;
          }
        }
      }

      if ($itemChanged) {
        // Element zum Schreiben sperren:
        $lockData = $itemService->GetLockData($item->ID);

        if (!$lockData)
          $itemService->Lock($item->ID);

        // Geändertes Element speichern:
        try {
          $itemService->Put($item);
        } catch (Exception $exception) {
          $error = $exception;
        }

        // Immer entsperren, auch wenn ein Fehler aufgetreten ist:
        try {
          if (!$lockData)
            $itemService->Unlock($item->ID);
        } catch (Exception $e) {
        }
      }
    }

    if($error)
      throw new GhAddonException('Datenübernahme fehlgeschlagen: ' . $error->getMessage(), $error->getCode());
    else
      return true;
  }

  /**
   * Ergänzt oder verändert ggf. eine Template-Variable bevor sie ausgegeben wird.
   *
   * @param string $var   Variablenname
   * @param string $value Aktueller Wert der Variablen
   * @param GhAddonOrder $order Auftrag
   * @param GhAddonCustomer $customer Kunde
   * @param GhAddonAccount $account Account
   * @param array $variables Addon-Variablen
   * @return string
   */
  protected function _processTemplateVariable($var, $value, $order, $customer, $account, $variables)
  {
    switch($var)
    {
      case 'Order.OrderNumber':
        $url = $this->_makeOrderLink($order, $customer, $account);

        if ($url) {
          $value = '<a href="' . $this->_addon->getBrowserLink($url) . '">' . $value . '</a>';
        } else {
          $handler = $this->_makeOrderOnClick($order, $customer, $account);

          if ($handler) {
            $value = '<a href="javascript:void();" onclick="' . $handler . '">' . $value . '</a>';
          }
        }
        break;

      case 'Order.CustomerNumber':
        $url = $this->_makeCustomerLink($order, $customer, $account);

        if ($url) {
          $value = '<a href="' . $this->_addon->getBrowserLink($url) . '">' . $value . '</a>';
        } else {
          $handler = $this->_makeCustomerOnClick($order, $customer, $account);

          if ($handler) {
            $value = '<a href="javascript:void();" onclick="' . $handler . '">' . $value . '</a>';
          }
        }
        break;

      case 'Order.InvoiceNumber':
        $url = $this->_makeInvoiceLink($order, $customer, $account);

        if($url) {
          $value = '<a href="' . $this->_addon->getBrowserLink($url) . '">' . $value . '</a>';
        } else {
          $handler = $this->_makeInvoiceOnClick($order, $customer, $account);

          if ($handler) {
            $value = '<a href="javascript:void();" onclick="' . $handler . '">' . $value . '</a>';
          }
        }
        break;

      case 'Shipping.Tracking.Code':
      case 'Shipping.Tracking.Codes':
      case 'Shipping.Tracking.URL':
      case 'Shipping.Tracking.URLs':
        $trackingInfo = array();

        foreach($order->Tracking as $tracking)
        {
          if($tracking['Url'])
            $trackingInfo[] = '<a href="' . $this->_addon->getBrowserLink($tracking['Url']) . '">' . ($tracking['Code'] ? $tracking['Code'] : 'Link') . '</a>';
          else
            $trackingInfo[] = $tracking['Code'];
        }

        $value = implode(' ', $trackingInfo);
        break;
      
      case 'Shipping.Tracking.Code1':
      case 'Shipping.Tracking.URL1':
        $trackingInfo = array();

        if(count($order->Tracking) >= 1)
        {
          $tracking = $order->Tracking[0];
          
          if($tracking['Url'])
            $trackingInfo[] = '<a href="' . $this->_addon->getBrowserLink($tracking['Url']) . '">' . ($tracking['Code'] ? $tracking['Code'] : 'Link') . '</a>';
          else
            $trackingInfo[] = $tracking['Code'];
        }

        $value = implode(' ', $trackingInfo);
        break;
      
      case 'Shipping.Tracking.Code2':
      case 'Shipping.Tracking.URL2':
        $trackingInfo = array();

        if(count($order->Tracking) >= 2)
        {
          $tracking = $order->Tracking[1];
          
          if($tracking['Url'])
            $trackingInfo[] = '<a href="' . $this->_addon->getBrowserLink($tracking['Url']) . '">' . ($tracking['Code'] ? $tracking['Code'] : 'Link') . '</a>';
          else
            $trackingInfo[] = $tracking['Code'];
        }

        $value = implode(' ', $trackingInfo);
        break;
      
      case 'Shipping.Tracking.Code3':
      case 'Shipping.Tracking.URL3':
        $trackingInfo = array();

        if(count($order->Tracking) >= 3)
        {
          $tracking = $order->Tracking[2];

          if($tracking['Url'])
            $trackingInfo[] = '<a href="' . $this->_addon->getBrowserLink($tracking['Url']) . '">' . ($tracking['Code'] ? $tracking['Code'] : 'Link') . '</a>';
          else
            $trackingInfo[] = $tracking['Code'];
        }

        $value = implode(' ', $trackingInfo);
        break;
  
      case 'Order.OpenAmount':
        if($value === null || $value === false)
          $value = '';
        elseif($value > 0.0)
          $value = '<span style="color:#800;font-weight:bold;">' . $value . '</span>';
        elseif($value < 0.0)
          $value = '<span style="color:#080;font-weight:bold;">' . $value . '</span>';
        else
          $value = '<span style="color:#080;">' . $value . '</span>';
        break;
    }

    return $value;
  }

  /**
   * Liefert einen Link, der eine URL im Webbrowser öffnet.
   *
   * @param string $url URL
   * @return string
   */
  protected function _makeBrowserLink($url)
  {
    return $this->_addon->getBrowserLink($url);
  }

  /**
   * Liefert einen Link, der eine Bestellung im Webbrowser oder einer Drittanwendung öffnet.
   *
   * @param GhAddonOrder $order Bestellung
   * @param GhAddonCustomer $customer Kunde
   * @param GhAddonAccount $account Account
   * @return string
   */
  protected function _makeOrderLink($order, $customer, $account)
  {
    return $account->getOrderUrl($order);
  }

  /**
   * Liefert einen onclick-Handler, der eine Bestellung im Webbrowser oder einer Drittanwendung öffnet.
   *
   * @param GhAddonOrder $order Bestellung
   * @param GhAddonCustomer $customer Kunde
   * @param GhAddonAccount $account Account
   * @return string
   */
  protected function _makeOrderOnClick($order, $customer, $account)
  {
    return $account->getOrderClickHandler($order);
  }

  /**
   * Liefert einen Link, der einen Kunden im Webbrowser oder einer Drittanwendung öffnet.
   *
   * @param GhAddonOrder $order Bestellung
   * @param GhAddonCustomer $customer Kunde
   * @param GhAddonAccount $account Account
   * @return string
   */
  protected function _makeCustomerLink($order, $customer, $account)
  {
    return $account->getCustomerUrl($order);
  }

  /**
   * Liefert einen onclick-Handler, der einen Kunden im Webbrowser oder einer Drittanwendung öffnet.
   *
   * @param GhAddonOrder $order Bestellung
   * @param GhAddonCustomer $customer Kunde
   * @param GhAddonAccount $account Account
   * @return string
   */
  protected function _makeCustomerOnClick($order, $customer, $account)
  {
    return $account->getCustomerClickHandler($order);
  }

  /**
   * Liefert einen Link, der eine Rechnung im Webbrowser oder einer Drittanwendung öffnet.
   *
   * @param GhAddonOrder $order Bestellung
   * @param GhAddonCustomer $customer Kunde
   * @param GhAddonAccount $account Account
   * @return string
   */
  protected function _makeInvoiceLink($order, $customer, $account)
  {
    return $account->getInvoiceUrl($order);
  }

  /**
   * Liefert einen onclick-Handler, der eine Rechnung im Webbrowser oder einer Drittanwendung öffnet.
   *
   * @param GhAddonOrder $order Bestellung
   * @param GhAddonCustomer $customer Kunde
   * @param GhAddonAccount $account Account
   * @return string
   */
  protected function _makeInvoiceOnClick($order, $customer, $account)
  {
    return $account->getInvoiceClickHandler($order);
  }

  /**
   * Startet die Verarbeitung eines eingehenden Requests und gibt das Ergebnis aus.
   */
  public function show()
  {
    try
    {
      $startTime = microtime(true);
      $this->_authenticate();
      @$result = json_encode($this->_processRequest());
      header('X-RenderTimeTotal: ' . round(microtime(true) - $startTime, 2) . ' s');
      echo '{"jsonrpc":"2.0","result":' . $result . ', "id": ' . $this->_rpcId . '}';
    }
    catch(Exception $e)
    {
      $json = '{"jsonrpc":"2.0","error":{"code":' . ((int)$e->getCode() ? (int)$e->getCode() : '500') . ',"message":' . json_encode($e->getMessage()) . ',"name":"' . get_class($e) . '"';

      if(get_class($e) == 'GhAddonEmptySearchException')
      {
        /* @var GhAddonEmptySearchException $e */
        $json .= ',"searchTerm":' . json_encode($e->getSearchTerm());
      }

      if($e instanceof GhAddonSettingsChangedException)
      {
        /* @var GhAddonSettingsChangedException $e */
        $json .= ',"settingsTimestamp":' . (int)$e->getSettingsTimestamp();
      }

      $json .= '}, "id": ' . $this->_rpcId . '}';
      echo $json;
    }
  }

  /**
   * Liefert einen gültigen Wert für ein benutzerdefiniertes Feld.
   *
   * @param mixed  $value Wert
   * @param int    $type  Typ
   * @param string $var   Name der Template-Variablen
   * @return mixed
   */
  protected function _getUserFieldValue($value, $type, $var)
  {
    switch($type)
    {
      case rpc_itemslib::ufdtDateTime:
        $timestamp = strtotime($value);

        if($timestamp)
          $value = date('Y-m-d H:i:s', $timestamp);
        else
          $value = '';
        break;

      case rpc_itemslib::ufdtFloat:
      case 7: //rpc_itemslib::ufdtCurrency:
        $value = str_replace(',', '.', $value);
        $value = preg_replace('/[^0-9\.]/s', '', $value);
        $value = sprintf('%f', $value);
        break;

      case rpc_itemslib::ufdtInteger:
      case rpc_itemslib::ufdtInt64:
        $value = preg_replace('/[^0-9]/s', '', $value);
        $value = sprintf('%d', $value);
        break;

      case rpc_itemslib::ufdtBoolean:
        $value = $value ? '-1' : '0';
        break;

      default:
        if(GhCoreStrings::endsWith($var, '.Sex.Number'))
        {
          if($value == rpc_itemslib::csMale)
            $value = 'Männlich';
          elseif($value == rpc_itemslib::csFemale)
            $value = 'Weiblich';
          else
            $value = 'Unbekannt';
        }
    }

    return $value;
  }

  /**
   * Liefert den Ordner in dem Cache-Dateien gespeichert werden.
   *
   * @return string
   */
  protected function _getCacheFolder()
  {
    return dirname(__FILE__) . '/cache/';
  }

  /**
   * Liest Daten aus dem Datei-Cache.
   * Falls zu dem Schlüssel keine Daten vorliegen, oder die Daten veraltet sind, wird null zurück geliefert.
   *
   * @throws GhAddonCacheLockedException falls der Cache-Eintrag gesperrt ist
   *
   * @param string $key Schlüssel, unter dem die Daten abgelegt sind
   * @param int $maxAge maximales Alter der Daten in Sekunden
   * @return string
   */
  protected function _readCache($key, $maxAge)
  {
    $path = $this->_getCacheFolder() . $key;

    if(!file_exists($path))
      return null;

    if(filemtime($path) < time() - $maxAge)
      return null;

    $file = fopen($path, 'r');

    if(!$file)
      return null;

    if(!flock($file, LOCK_SH))
    {
      fclose($file);
      throw new GhAddonCacheLockedException('Failed to obtain read lock');
    }

    $content = fread($file, filesize($path));
    flock($file, LOCK_UN);
    fclose($file);

    return $content;
  }

  /**
   * Beginnt eine Schreiboperation auf den Cache.
   * Diese Operation sperrt die Cache-Datei. Die Schreiboperation muss zwingend mit endCacheWrite und dem Rückgabewert
   * von beginCacheWrite als Parameter beendet werden, ansonsten bleibt die Datei gesperrt.
   *
   * @throws GhAddonCacheLockedException falls die Cache-Datei nicht zum Schreiben gesperrt werden konnte
   *
   * @param string $key Cache-Schlüssel unter dem Daten gespeichert werden sollen
   * @return resource
   */
  protected function _beginCacheWrite($key)
  {
    $path = $this->_getCacheFolder();

    if(!is_dir($path))
      mkdir($path, 0777, true);

    $path .= $key;

    $file = fopen($path, 'c');

    if(!$file || !flock($file, LOCK_EX))
      throw new GhAddonCacheLockedException('Failed to obtain write lock');

    ftruncate($file, 0);

    $this->_cache = $file;
  }

  /**
   * Beendet eine Schreiboperation auf den Cache.
   *
   * @param resource $file Cache-Datei Ressource
   * @param string $data Daten, die in den Cache geschrieben werden sollen
   */
  protected function _endCacheWrite($file, $data)
  {
    fwrite($file, $data);

    flock($file, LOCK_UN);
    fclose($file);
  }

  /**
   * Löscht Cache-Dateien, die ein bestimmtes Mindestalter haben.
   *
   * @param int $minAge minimales Alter in Sekunden
   */
  protected function _clearCache($minAge = 0)
  {
    $path = $this->_getCacheFolder();

    if(!is_dir($path))
      return;

    $maxTime = time() - $minAge;
    $dir = opendir($path);

    try
    {
      $file = readdir($dir);

      while($file !== false)
      {
        if($file != '.' && $file != '..')
        {
          if(filemtime($path . $file) <= $maxTime)
            unlink($path . $file);
        }

        $file = readdir($dir);
      }
    }
    catch(Exception $exception)
    {
    }

    closedir($dir);
  }

  /**
   * Sucht nach einem Kontakt anhand der Kunden-ID, E-Mail-Adresse oder Telefonnummer.
   *
   * @param array $params Parameter
   * @return array
   */
  protected function _getContactRef($params)
  {
    $email = GhAddonData::safeGet($params, '0/email');
    $phone = GhAddonData::safeGet($params, '0/phone');
    $customerId = GhAddonData::safeGet($params, '0/customerId');

    if($phone)
      $phone = GhCoreLocale::fixPhoneNumber($phone);

    $result = array('email' => $email, 'phone' => $phone, 'customerId' => $customerId, 'contactRef' => 0);

    if($customerId)
    {
      $userFieldId = $this->_addon->getAddonSetting('CustomerIDUserField', 0);

      if($userFieldId)
      {
        $field = rpc_itemslib::newRpcUserFieldItem();
        $field->FieldRef = $userFieldId;
        $field->Value = $customerId;

        $filter = rpc_itemslib::newRpcItemFilter();
        $filter->Kinds = GhRpcUtils::newSet(rpc_itemslib::ikContact);
        $filter->UserFields = array($field);
        $items = rpc_itemslib::newRpcItems($this->_addon->getGreyhoundClient())->GetList($filter, rpc_itemslib::iobCreated, rpc_itemslib::iodAscending, 0, 0, array(), 0);
        $result['contactRef'] = GhAddonData::safeGet($items, '0/ID', 'int');
      }
    }

    if(!$result['contactRef']) {
      $search = array();

      if($email)
        $search[] = $email;

      if($phone)
        $search[] = $phone;

      $result['search'] = implode(' | ', $search);

      $filter = rpc_itemslib::newRpcItemFilter();
      $filter->Kinds = GhRpcUtils::newSet(rpc_itemslib::ikContact);
      $filter->SearchFor = $result['search'];
      $filter->SearchFields = GhRpcUtils::newSet(rpc_itemslib::isfAddress);

      if($filter->SearchFor)
      {
        $items = rpc_itemslib::newRpcItems($this->_addon->getGreyhoundClient())->GetList($filter, rpc_itemslib::iobCreated, rpc_itemslib::iodAscending, 0, 0, array(), 0);
        $result['contactRef'] = GhAddonData::safeGet($items, '0/ID', 'int');
      }
    }

    return $result;
  }

  /**
   * Liefert die Daten zum Versenden des Debug-Logs zurück, sofern eine Debug-Datei vorliegt.
   *
   * @throws GhAddonException falls kein Debug-Log vorliegt
   *
   * @param array $params Parameter
   * @return array
   */
  protected function _sendDebugLog($params)
  {
    $debugLog = $this->_addon->getDebugLogFile();

    if(!$this->_addon->getAddonSetting('DebugMode', false) || !$debugLog || !file_exists($debugLog)) {
      $e = new GhAddonException('Es liegt zurzeit kein Debug-Log vor.');
      $e->dontLog = true;
      throw $e;
    }

    return array(
      'email' => $this->_addon->getMaintainerEmail(),
      'subject' => 'Debug-Log für ' . $this->_addon->getName() . ' ' . $this->_addon->getVersion() . ' (build ' . $this->_addon->getBuild() . ')',
      'attachments' => array($debugLog)
    );
  }
}
