<?php
/*
 * GREYHOUND Addon Framework
 *
 * Copyright (c) 2011-2019 GREYHOUND Software GmbH & Co. KG
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of the GREYHOUND Addon Framework and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including without
 * limitation the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to permit persons to
 * whom the Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/**
 * Datenklasse für das Ergebnis einer Datenabfrage des Addons beim Drittsystem.
 *
 * @package GhAddonFramework
 * @subpackage core
 * @category PHP Addon Framework
 *
 * @author digital guru GmbH &amp; Co. KG <develop@greyhound-software.com>
 * @copyright 2011-2019 digital guru GmbH &amp; Co. KG
 * @link greyhound-software.com
 */
abstract class GhAddonData
{
  /**
   * Formatangabe ohne festgelegtes Format.
   * @var string
   */
  const FORMAT_NONE = '';

  /**
   * Formatangabe für automatisches Format (HTML/Text).
   * @var string
   */
  const FORMAT_AUTO = 'auto';

  /**
   * Formatangabe für HTML.
   * @var string
   */
  const FORMAT_HTML = 'html';

  /**
   * Formatangabe für Text.
   * @var string
   */
  const FORMAT_TEXT = 'text';

  /**
   * Kunden- und Auftragsdaten (Addon-spezifisch).
   * @var mixed
   */
  protected $_data;

  /**
   * Die ausgewählte Bestellung (falls nach einer bestimmten ID gesucht wurde).
   * @var int
   */
  protected $_selectedOrderId;

  /**
   * Suchbegriff, der zum Ermitteln dieser Daten verwendet wurde.
   * @var string
   */
  protected $_searchTerm;

  /**
   * Diese Methode muss überladen werden und alle Kunden dieser Abfrage liefern.
   *
   * @return GhAddonCustomer[]
   */
  abstract public function getCustomers();

  /**
   * Diese Methode muss überladen werden und alle Bestellungen dieser Abfrage
   * liefern.
   *
   * @param string $customerId Kunden-ID, falls nur Bestellungen dieses Kunden
   *   geliefert werden sollen
   * @return GhAddonOrder[]
   */
  abstract public function getOrders($customerId = null);

  /**
   * Liefert die Kunden- und Auftragsdaten (Addon-spezifisch).
   *
   * @return mixed
   */
  public function getData()
  {
    return $this->_data;
  }

  /**
   * Setzt die Kunden- und Auftragsdaten (Addon-spezifisch).
   *
   * @param $data
   */
  public function setData($data)
  {
    $this->_data = $data;
  }

  /**
   * Setzt die ausgewählte Bestell-ID.
   * @param int $orderId Bestell-ID
   */
  public function setSelectedOrderId($orderId)
  {
    $this->_selectedOrderId = $orderId;
  }

  /**
   * Liefert die ausgewählte Bestell-ID.
   *
   * @return int
   */
  public function getSelectedOrderId()
  {
    return $this->_selectedOrderId;
  }

  /**
   * Setzt den Suchbegriff, der für das Ermitteln dieser Daten verwendet wurde.
   *
   * @param string $searchTerm Suchbegriff
   */
  public function setSearchTerm($searchTerm)
  {
    $this->_searchTerm = $searchTerm;
  }

  /**
   * Liefert den Suchbegriff, der für das Ermitteln dieser Daten verwendet wurde.
   *
   * @return string
   */
  public function getSearchTerm()
  {
    return $this->_searchTerm;
  }

  /**
   * Liefert einen Kunden anhand seiner ID.
   *
   * @param string $id ID des Kunden
   * @return GhAddonCustomer
   */
  public function getCustomer($id)
  {
    foreach($this->getCustomers() as $customer)
      if($customer->ID == $id)
        return $customer;

    return null;
  }

  /**
   * Liefert eine Bestellung anhand ihrer ID.
   *
   * @param string $id ID der Bestellung
   * @return GhAddonOrder
   */
  public function getOrder($id)
  {
    foreach($this->getOrders() as $order)
      if($order->ID == $id)
        return $order;

    return null;
  }

  /**
   * Liefert die Bestellungen die für einen Kunden vorliegen.
   *
   * @param GhAddonCustomer $customerId ID des Kunden
   * @return GhAddonOrder[]
   */
  public function getOrdersByCustomer($customerId)
  {
    $orders = array();

    foreach($this->getOrders() as $order)
      if($order->CustomerID == $customerId)
        $orders[] = $order;

    return $orders;
  }

  /**
   * Liefert einen Wert aus einem Ergebnisobjekt (oder Array) anhand eines Daten-Pfades.
   * Wird ein gewünschter Datentyp angegeben, so wird das Ergebnis in diesen Typ
   * umgewandelt.
   *
   * @param object|array $object    Ergebnisobjekt oder Daten-Array
   * @param string       $path      Daten-Pfad (durch '/' getrennt)
   * @param string       $type      gewünschter Typ (int, double, boolean, string, array, object)
   * @param boolean      $allowNull true: null als Ergebnis zulassen, andernfalls wird ein leerer Wert (leeres Objekt, Array) zurückgegeben
   * @param mixed        $defaultValue Vorgabewert, falls der Wert nicht gefunden wird
   * @return mixed
   */
  public static function safeGet($object, $path, $type = 'string', $allowNull = false, $defaultValue = null)
  {
    if (!is_object($object) && !is_array($object)) {
      return $allowNull ? null : self::convertValue(null, $type);
    }

    $parts = explode('/', trim($path, '/'));

    if (empty($parts)) {
      return self::convertValue($object, $type);
    }

    $current = $object;

    foreach ($parts as $part) {
      if (is_object($current) && isset($current->$part)) {
        $current = $current->$part;
      } elseif (is_array($current) && isset($current[$part])) {
        $current = $current[$part];
      } elseif ($defaultValue !== null) {
        return $defaultValue;
      } else {
        return $allowNull ? null : self::convertValue(null, $type);
      }
    }

    return self::convertValue($current, $type);
  }

  /**
   * Liefert einen Wert aus einem Ergebnisobjekt (oder Array) anhand eines Daten-Pfades als Array zurück.
   * Befindet sich an der angegebenen Position ein Array, so wird dieser geliefert, andernfalls ein Array mit dem
   * Wert, der sich an der Position befindet.
   *
   * @param object|array $object Ergebnisobjekt oder Daten-Array
   * @param string       $path   Daten-Pfad (durch '/' getrennt)
   * @return array
   */
  public static function getArray($object, $path = '')
  {
    $parts = explode('/', trim($path, '/'));

    if(strlen($path) == 0 || empty($parts))
    {
      if(is_array($object))
        return $object;
      elseif($object)
        return array($object);
      else
        return array();
    }

    $current = $object;

    foreach($parts as $part)
    {
      if(is_object($current) && isset($current->$part))
        $current = $current->$part;
      elseif(is_array($current) && isset($current[$part]))
        $current = $current[$part];
      else return array();
    }

    if(is_array($current))
      return $current;
    elseif($current)
      return array($current);
    else
      return array();
  }

  /**
   * Wandelt Daten in einen gewünschten Typ um.
   *
   * @param mixed $value Wert
   * @param string $type Typ (int, double, boolean, string, array, object)
   * @return mixed
   */
  public static function convertValue($value, $type)
  {
    if(($value instanceof SimpleXmlElement) && $type != 'array' && $type != 'object')
      $value = (string)$value;

    if(!is_scalar($value))
    {
      switch($type)
      {
        case 'array':
          if(is_array($value))
            return $value;
          elseif(is_object($value))
            return get_object_vars($value);
          else
            return array();

        case 'object':
          if(is_object($value))
            return $value;
          elseif(is_array($value))
            return (object)$value;
          else
            return new stdClass();

        case 'unchanged':
          return $value;

        default:
          $value = null;
      }
    }

    switch($type)
    {
      case 'int':
        return (int)$value;

      case 'double':
      case 'float':
        return (double)$value;

      case 'boolean':
        return $value ? true : false;
      
      case 'unchanged':
        return $value;

      default:
        return (string)$value;
    }
  }

  /**
   * Liefert die Prüfsumme für eine GREYHOUND Vorgangsnummer.
   *
   * @param int $number number
   * @return int
   */
  public static function getItemNumberChecksum($number)
  {
    $str = strtoupper((string)$number);
    $checksum = 0;

    for($i=0; $i<strlen($str); $i++)
    {
      $value = ord($str[$i]);

      if($value >= 48 && $value <= 57)
        $value -= 48;
      elseif($value >= 65 && $value <= 90)
        $value -= 55;

      $checksum += $value;
    }

    $checksum %= 100;

    return $checksum;
  }

  /**
   * Vergleichsfunktion für Bestellungen.
   *
   * Bestellungen werden absteigend nach dem Bestelldatum sortiert.
   * Bei gleichem Bestelldatum wird absteigend nach der OrderID sortiert.
   *
   * Das Ergebnis ist -1 falls die erste Bestellung neuer als die Zweite ist,
   * 1 falls sie älter als die Zweite ist, oder 0 falls beide gleich neu sind.
   *
   * @param GhAddonOrder $order1 erste Bestellung
   * @param GhAddonOrder $order2 zweite Bestellung
   * @return int
   */
  public function compareOrders($order1, $order2)
  {
    $time1 = $order1->Date instanceof GhRpcDateTime ? $order1->Date->getTimestamp() : 0;
    $time2 = $order2->Date instanceof GhRpcDateTime ? $order2->Date->getTimestamp() : 0;

    if($time1 > $time2)
      return -1;
    elseif($time1 < $time2)
      return 1;

    if($order1->ID > $order2->ID)
      return -1;
    elseif($order1->ID < $order2->ID)
      return 1;
    else
      return 0;
  }

  /**
   * Vergleichsfunktion für Kunden.
   *
   * Kunden werden alphabetisch aufsteigend nach Name sortiert.
   * Das Ergebnis ist -1 falls der erste Kunde alphabetisch vor dem zweiten
   * kommt, 1 falls er nach dem zweiten kommt, oder 0 falls beide den gleichen
   * Namen haben.
   *
   * @param GhAddonCustomer $customer1 erster Kunde
   * @param GhAddonCustomer $customer2 zweiter Kunde
   * @return int
   */
  public function compareCustomers($customer1, $customer2)
  {
    // Umlaute bei der Sortierung wie normale Buchstaben behandeln:
    $name1 = str_replace(array('ä', 'ö', 'ü', 'Ä', 'Ö', 'Ü', 'ß'), array('a', 'o', 'u', 'A', 'O', 'U', 's'), $customer1->Name);
    $name2 = str_replace(array('ä', 'ö', 'ü', 'Ä', 'Ö', 'Ü', 'ß'), array('a', 'o', 'u', 'A', 'O', 'U', 's'), $customer2->Name);

    if($name1 < $name2)
      return -1;
    elseif($name1 > $name2)
      return 1;
    else
      return 0;
  }

  /**
   * Liefert eine Tracking-URL für einen Versanddienstleister und eine Sendungsnummer.
   *
   * @param string $trackingId       Sendungsnummer
   * @param string $shippingProvider Versanddienstleister (Kürzel, z.B. DHL, UPS, DPD)
   * @param string $zipCode          Postleitzahl
   * @return string
   */
  public static function getTrackingUrl($trackingId, $shippingProvider, $zipCode = null)
  {
    if(!$trackingId)
      return null;

    switch(trim(strtoupper($shippingProvider)))
    {
      case 'DHL':
      case 'DPI':
        return 'http://nolp.dhl.de/nextt-online-public/track.do?idc=' . $trackingId . '&lang=de';

      case 'DPD':
        return 'https://tracking.dpd.de/cgi-bin/delistrack?typ=1&lang=de&pknr=' . $trackingId;

      case 'FEDEX':
        return 'http://www.fedex.com/Tracking?tracknumbers=' . $trackingId;

      case 'GLS':
        return 'http://www.gls-germany.com/online/paketstatus.php3?paketliste[0]=' . $trackingId;

      case 'HERMES':
      case 'HVS':
        return 'https://www.myhermes.de/wps/portal/paket/SISYR?auftragsNummer=' . $trackingId;

      case 'UPS':
        return 'http://wwwapps.ups.com/ietracking/tracking.cgi?tracknum=' . $trackingId . '&IATA=de&Lang=ger';

      case 'USPS':
        return 'http://trkcnfrm1.smi.usps.com/PTSInternetWeb/InterLabelInquiry.do?origTrackNum=' . $trackingId;
    }

    return null;
  }

  /**
   * Wandelt ein Datum aus der API in ein GhRpcDateTime Objekt um (oder null, falls es sich nicht um ein gültiges Datum handelt).
   * 
   * @param string $dateStr Datums-String
   * @return GhRpcDateTime
   */
  public static function getDate($dateStr)
  {
    $dateStr = strval($dateStr);

    if(is_numeric($dateStr)) {
      return new GhRpcDateTime($dateStr);
    }
    elseif(strlen($dateStr) >= 19) {
      return new GhRpcDateTime($dateStr);
    }
    elseif(strlen($dateStr) == 10) {
      return new GhRpcDateTime($dateStr . ' 00:00:00');
    }
    else {
      return null;
    }
  }

  /**
   * Schneidet die Nachkommastellen einer Fließkommazahl ab falls sie 0 sind.
   *
   * @param float $number Zahl
   * @param int $decimals Dezimalstellen (null = unbegrenzt)
   * @return int
   */
  protected function _formatFloat($number, $decimals = null)
  {
    if((int)$number == $number || $decimals === 0)
      return (int)$number;
    elseif($decimals !== null)
      return sprintf('%.' . intval($decimals) . 'f', $number);
    else
      return $number;
  }

  /**
   * Wandelt ein Datum aus der API in ein GhRpcDateTime Objekt um (oder null, falls es sich nicht um ein gültiges Datum handelt).
   * 
   * @param string $dateStr Datums-String
   * @return GhRpcDateTime
   */
  protected function _getDate($dateStr)
  {
    return self::getDate($dateStr);
  }

  /**
   * Liefert die Kundennummer eines Auftrags, ggf. in einem Anchor-Tag mit Hyperlink auf den Kunden im Drittsystem.
   * 
   * @var GhAddonOrder $order Auftrag
   * @return string
   */
  protected function _getCustomerNumberHtml($order)
  {
    if (!$order->CustomerNumber) {
      return null;
    }

    $url = $this->_account->getCustomerUrl($order);
    $onClick = $this->_account->getCustomerClickHandler($order);

    if ($url) {
      return '<a href=\'' . $this->_addon->getBrowserLink($url) . '\'>' . $order->CustomerNumber . '</a>';
    } elseif ($onClick) {
      return '<a href="javascript:void(0);" onclick="' . $onClick . '">' . $order->CustomerNumber . '</a>';
    } else {
      return $order->CustomerNumber;
    }
  }

  /**
   * Liefert die Auftragsnummer eines Auftrags, ggf. in einem Anchor-Tag mit Hyperlink auf den Auftrag im Drittsystem.
   * 
   * @var GhAddonOrder $order Auftrag
   * @return string
   */
  protected function _getOrderNumberHtml($order)
  {
    if (!$order->OrderNumber) {
      return null;
    }

    $url = $this->_account->getOrderUrl($order);
    $onClick = $this->_account->getOrderClickHandler($order);

    if ($url) {
      return '<a href=\'' . $this->_addon->getBrowserLink($url) . '\'>' . $order->OrderNumber . '</a>';
    } elseif ($onClick) {
      return '<a href="javascript:void(0);" onclick="' . $onClick . '">' . $order->OrderNumber . '</a>';
    } else {
      return $order->OrderNumber;
    }
  }

  /**
   * Liefert die Rechnungsnummer eines Auftrags, ggf. in einem Anchor-Tag mit Hyperlink auf die Rechnung im Drittsystem.
   * 
   * @var GhAddonOrder $order Auftrag
   * @return string
   */
  protected function _getInvoiceNumberHtml($order)
  {
    if (!$order->InvoiceNumber) {
      return null;
    }

    $url = $this->_account->getInvoiceUrl($order);
    $onClick = $this->_account->getInvoiceClickHandler($order);

    if ($url) {
      return '<a href=\'' . $this->_addon->getBrowserLink($url) . '\'>' . $order->InvoiceNumber . '</a>';
    } elseif ($onClick) {
      return '<a href="javascript:void(0);" onclick="' . $onClick . '">' . $order->InvoiceNumber . '</a>';
    } else {
      return $order->InvoiceNumber;
    }
  }

  /**
   * Liefert die Artikelnummer einer Auftragsposition, ggf. in einem Anchor-Tag mit Hyperlink auf den Artikel im Drittsystem.
   * 
   * @var GhAddonOrderItem $orderItem Auftragsposition
   * @var GhAddonOrder $order Auftrag
   * @return string
   */
  protected function _getProductNumberHtml($orderItem, $order = null)
  {
    if (!$orderItem->ItemNumber) {
      return null;
    }

    $url = $this->_account->getProductUrl($orderItem, $order);
    $onClick = $this->_account->getProductClickHandler($orderItem, $order);

    if ($url) {
      return '<a href=\'' . $this->_addon->getBrowserLink($url) . '\'>' . $orderItem->ItemNumber . '</a>';
    } elseif ($onClick) {
      return '<a href="javascript:void(0);" onclick="' . $onClick . '">' . $orderItem->ItemNumber . '</a>';
    } else {
      return $orderItem->ItemNumber;
    }
  }

  /**
   * Liefert die Zahlungs-Transaktions-ID eines Auftrags, ggf. in einem Anchor-Tag mit Hyperlink auf die Transaktion beim Zahlungsanbieter.
   * 
   * @var GhAddonOrder $order Auftrag
   * @return string
   */
  protected function _getPaymentTransactionHtml($order)
  {
    if (!$order->PaymentTransactionID) {
      return null;
    }

    $url = $this->_addon->getPaymentTransactionUrl($order->PaymentTransactionID, $order->PaymentMethod);

    if ($url) {
      return '<a href=\'' . $this->_addon->getBrowserLink($url) . '\'>' . $order->PaymentTransactionID . '</a>';
    } else {
      return $order->PaymentTransactionID;
    }
  }
}